#!/usr/bin/env python

import getopt
import os
import sys
import datetime
import re

try:
    # load from system
    import yaml
except ImportError:
    #check load yaml-module from STP-library
    prog_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0,  prog_dir + '/../libpy/lib/')
    try:
        import yaml
    except ImportError:
        print ('yaml-module not found in system modules and STP-library')
        sys.exit(1)


# dictionary for command line arguments
RESULTS_ROOT_DIR = 'path_to_results'
INFO_OUTPUT_MODE = 'output_mode'
DATE_START = 'date_start'
DATE_END = 'date_end'

TESTING_NAME = 'testing_name'
VK_NAME = 'VK_name'
TEST_NAME = 'test_name'

TESTING_FLAG = 'testing_flag'
VK_FLAG = 'vk_flag'
TEST_FLAG = 'test_flag'

VERBOSE = 'verbose'


def print_splitter(ch, count):
    for i in range(count):
        sys.stdout.write(ch)
    sys.stdout.write('\n')


def print_total_time(sec):
    day = sec / (24 * 60 * 60)
    sec -= day * (24 * 60 * 60)
    hour = sec / (60 * 60)
    sec -= hour * (60 * 60)
    min = sec / 60
    sec -= min * 60

    return "%dd %dh %dm %ds" % (day, hour, min, sec)


def print_test_info(args_list, test_name, test_prop, prefix=''):
    info_mode = 'short'
    if INFO_OUTPUT_MODE in args_list:
        info_mode = args_list[INFO_OUTPUT_MODE]

    if info_mode == 'medium':
        sys.stdout.write(prefix + '%s' % (test_name))
        sys.stdout.write(' [passes:%d errors:%d]\n' %
                         (test_prop['passes'], test_prop['errors']))
    elif info_mode == 'full':
        sys.stdout.write(prefix + '%s' % (test_name))
        try:
            test_type = '%s ' % (test_prop['type'])
        except KeyError:
            test_type =''
        try:
            test_count = 'amount:%d ' % (test_prop['amount'])
        except KeyError:
            test_count =''
        sys.stdout.write(' [%s%spasses:%d errors:%d time:%d]\n' %
                         (test_type, test_count,
                          test_prop['passes'], test_prop['errors'], test_prop['time']))


def print_vk_info(args_list, vk_name, vk_prop, prefix=''):
    info_mode = 'short'
    tab = prefix
    if INFO_OUTPUT_MODE in args_list:
        info_mode = args_list[INFO_OUTPUT_MODE]

    if info_mode == 'medium':
        sys.stdout.write(prefix + '%s [%s]\n' % (vk_name, vk_prop['resume']))

    if info_mode == 'full':
        print(prefix + vk_name)
        tab = prefix + '|\t'
        print(tab + 'resume:'.ljust(10) + vk_prop['resume'])
        print(tab + 'time:'.ljust(10) + print_total_time(vk_prop['time']))

    tab += '\t'
    if vk_prop['tests']:
        for (test_name, test_prop) in vk_prop['tests'].items():
            if TEST_NAME in args_list:
                if args_list[TEST_NAME] != test_name:
                    continue
            print_test_info(args_list, test_name, test_prop, tab)
    else:
        print tab + 'test [None]'

    print(prefix + '\\'.ljust(4, '_'))


def print_testing_info(args_list, key, run_test):
    max_width = 80

    # testing title
    name_w = 32
    cell_w = (max_width - name_w) / 2
    date = datetime.datetime.strptime(run_test['date'], '%Y-%m-%d %H:%M:%S')
    sys.stdout.write('%*.*s%*.*s%*.*s\n' % (name_w, name_w, 'Name: %s' % key,
                     cell_w, cell_w, run_test['resume'].upper().center(cell_w),
                     cell_w, cell_w, '%02d.%02d.%04d ~ %02d:%02d:%02d' %
                                            (date.day, date.month, date.year,
                                             date.hour, date.minute,
                                             date.second)))

    if not INFO_OUTPUT_MODE in args_list:
        info_mode = 'short'
    else:
        info_mode = args_list[INFO_OUTPUT_MODE]

    if info_mode == 'short':
        success_vk = 0;
        for vk in run_test['vk'].values():
            if vk['resume'].lower() == 'success':
                success_vk += 1;
        print '\t%d of %d VK successfully tested' % (success_vk,
                                                     len(run_test['vk']))
        return
    if info_mode == 'medium' or info_mode == 'full':
        print '\tduration ' + print_total_time(run_test['duration'])
        print '\tfolder at host .../' + run_test['path']

    if info_mode == 'full':
        print '\t' + run_test['report']

    if run_test['vk']:
        for (vkname, vk) in run_test['vk'].items():
            if VK_NAME in args_list:
                if args_list[VK_NAME] != vkname:
                    continue

                    if VK_FLAG in args_list:
                        if args_list[VK_FLAG] != vk['resume'].lower():
                            continue
            print_vk_info(args_list, vkname, vk, '')
    else:
        print '\'none\' [no result]'


def pars_file(args_lst, path):
    name = path.split('/')[-1]
    res_start = False
    res_end = False

    if VERBOSE in args_lst:
        print '\n'
        print 'File name: ' + os.path.abspath(path)
        print '-----------' + '-'*len(os.path.abspath(path))

    file_data = datetime.datetime.strptime('%s%s%s' % (name[:4], name[5:7], name[8:10]), '%Y%m%d')

    if DATE_START in args_lst:
        if args_lst[DATE_START] <= file_data:
            res_start = True;
    else:
        res_start = True

    if DATE_END in args_lst:
        if args_lst[DATE_END] >= file_data:
            res_end = True
    else :
        res_end = True

    if res_start and res_end:
        try:
            results = yaml.load(open(path))
        except (OSError, IOError), err:
            print (err)
            return
        except yaml.YAMLError, err:
            #print (err)
            print ('YAML-file structure is incorrect')
            if hasattr(err, 'problem_mark'):
                mark = err.problem_mark
                print ("Error position: (%s:%s)" % (mark.line+1, mark.column+1))
            print ('Check result files')
            return

        # loop by testing
        for (run_test_name, run_test) in results.items():
            run_test_name = str(run_test_name) # explicit type
            if TESTING_NAME in args_lst:
                if run_test_name != args_lst[TESTING_NAME]:
                    continue

            if TESTING_FLAG in args_lst:
                if args_lst[TESTING_FLAG] != run_test['resume'].lower():
                    continue

            monitorkey = ' monitor'
            if monitorkey in run_test_name and run_test_name[:-len(monitorkey)] in results:
                continue

            print_testing_info(args_lst, run_test_name, run_test)


def file_is_results(full_path):
    if (not os.path.isfile(full_path)):
        return False
    name = full_path.split('/')[-1]
    p = re.compile('^[\d]{4}_(0[1-9]|1[0-2])_(0[1-9]|1[0-9]|2[0-9]|3[01])_[\d]{2}.res$')
    return not p.match(name) is None


def check_folder(args_lst, path):
    try:
        if os.path.isfile(path):
            if file_is_results(path):
                pars_file(args_lst, path)
                return
    except OSError, err:
        pass

    try:
        #print 'path:\t' + path
        file_lst = os.listdir(path)
        #print file_lst
    except OSError, err:
        print (str(err) + '\nCheck path to results folder')
        return

    for name in file_lst:
        full_name = path + '/' + name
        if os.path.isdir(full_name):
            if name.startswith('.'):
                continue
            check_folder(args_lst, full_name)
        elif file_is_results(full_name):
            pars_file(args_lst, full_name)


def rv_usage():
    print 'Usage: res_viewer.py [-husmf] [-r path] [-S dd.mm.yyyy] [-E dd.mm.yyyy] [-n NAME]' \
          ' [-v NAME] [-t NAME]'


def rv_help():
    print 'Help:\n' \
          '  -h, --help\t\t\t\t  help\n' \
          '  -r PATH, --results=PATH\t\t  path to directory with results\n' \
          '  -s, --short-info\t\t\t  print results in short info mode\n' \
          '  -m, --medium-info\t\t\t  print results in medium info mode\n' \
          '  -f, --full-info\t\t\t  print results in full info mode\n' \
          '  -S DD.MM.YYYY, --date-start=DD.MM.YYYY  date of start testing (for search)\n' \
          '  -E DD.MM.YYYY, --date-end=DD.MM.YYYY\t  date of end testing (for search)\n' \
          '  -n NAME, --name=NAME\t\t\t  name testing (for search)\n' \
          '  -v NAME, --vk=NAME\t\t\t  name of vk for search\n' \
          '  -t NAME, --test=NAME\t\t\t  name of test for search\n' \
          '  --success-testing\t\t\t  search only testing that successfully completed\n' \
          '  --unsuccess-testing\t\t\t  search only testing, which ended in failure\n' \
          '  --success-vk\t\t\t\t  search only those VK that testing was successful\n' \
          '  --unsuccess-vk\t\t\t  search only those computers that failed testing\n'\
          '  --success-test\t\t\t  search only successful tests\n' \
          '  --unsuccess-test\t\t\t  search only unsuccessful tests\n' \
          '  -u, --usage\t\t\t\t  usage\n' \
          '  --verbose\t\t\t\t  verbose additional information'


def pars_args():
    arg_list = {};
    short_opt = 'hr:smfS:E:n:v:t:u'
    long_opt = 'help results= short-info medium-info full-info name= ' \
               'vk= test= date-start= date-end= ' \
               'success-testing success-vk success-test ' \
               'unsuccess-testing unsuccess-vk unsuccess-test ' \
               'usage verbose'
    try:
        optlist, arg = getopt.getopt(sys.argv[1:], short_opt, long_opt.split())
    except getopt.GetoptError, err:
        print err
        rv_usage()
        rv_help()
        sys.exit()

    date_fmt = '%d.%m.%Y'

    for o, v in optlist:
        if o in ('-h', '--help'):
            rv_usage()
            rv_help()
            sys.exit(0)
        elif o in ('-r', '--results'):
            arg_list[RESULTS_ROOT_DIR] = v
        elif o in ('-s', '--short-info'):
            arg_list[INFO_OUTPUT_MODE] = 'short'
        elif o in ('-m', '--medium-info'):
            arg_list[INFO_OUTPUT_MODE] = 'medium'
        elif o in ('-f', '--full-info'):
            arg_list[INFO_OUTPUT_MODE] = 'full'
        elif o in ('-n', '--name'):
            arg_list[TESTING_NAME] = v
        elif o in ('-v', '--vk'):
            arg_list[VK_NAME] = v
        elif o in ('-t', '--test'):
            arg_list[TEST_NAME] = v
        elif o in ('-S', '--date-start'):
            date_s = datetime.datetime.strptime(v, date_fmt)
            arg_list[DATE_START] = date_s
        elif o in ('-E', '--date-end'):
            date_e = datetime.datetime.strptime(v, date_fmt)
            arg_list[DATE_END] = date_e
        elif o == '--success-testing':
            arg_list[TESTING_FLAG] = 'success'
        elif o == '--success-vk':
            arg_list[VK_FLAG] = 'success'
        elif o == '--success-test':
            arg_list[TEST_FLAG] = 'success'
        elif o == '--unsuccess-testing':
            arg_list[TESTING_FLAG] = 'unsuccess'
        elif o == '--unsuccess-vk':
            arg_list[VK_FLAG] = 'unsuccess'
        elif o == '--unsuccess-test':
            arg_list[TEST_FLAG] = 'unsuccess'
        elif o in ('-u', '--usage'):
            rv_usage()
            sys.exit(0)
        elif o == '--verbose':
            arg_list[VERBOSE] = 'verbose'
        else:
            rv_usage()
            rv_help()
            assert False, 'unknown args'

    return arg_list


def pars_results():
    arg_list = pars_args()

    if RESULTS_ROOT_DIR in arg_list:
        check_folder(arg_list, arg_list[RESULTS_ROOT_DIR])
    else :
        check_folder(arg_list, prog_dir + '/../results/')


if __name__ == '__main__':
    pars_results()
