#! /usr/bin/env python

import optparse
import os
import logging
import re
import datetime
import sys

logging.basicConfig(
    format='%(levelname)s %(asctime)s %(filename)s %(lineno)d %(message)s')
log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)

try:
    import yaml
except ImportError:
    #check load yaml-module from STP-library
    prog_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0,  prog_dir + '/../libpy/lib/')
    try:
        import yaml
    except ImportError:
        log.error('yaml-module not found in system modules and STP-library')
        sys.exit(1)


DEFAULT_RESULTS_PATH = './../results'
vk_mode_list = ['all', 'success', 'unsuccess', 'list_names', 'list_ip']
test_mode_list = ['all', 'success', 'unsuccess', 'list_names', 'list_types']

def date_from_str(date_str, date_fmt='%d.%m.%Y'):
    d = None
    try:
        d = datetime.datetime.strptime(date_str, date_fmt).date()
    except (TypeError, ValueError), err:
        raise err
    return d


def vararg_callback(option, opt_str, value, parser):
    assert value is None
    value = []

    def floatable(str):
        try:
            float(str)
            return True
        except ValueError:
            return False

    for arg in parser.rargs:
        # stop on --foo like options
        if arg[:2] == "--" and len(arg) > 2:
            break
        # stop on -a, but not on -3 or -3.0
        if arg[:1] == "-" and len(arg) > 1 and not floatable(arg):
            break
        value.append(arg)

    del parser.rargs[:len(value)]
    setattr(parser.values, option.dest, value)


def date_callback(option, opt_str, value, parser):
    # assert value is None
    try:
        d = date_from_str(value)
    except (TypeError, ValueError):
        parser.error('Date specified in the wrong format')
    else:
        setattr(parser.values, option.dest, d)


def opt_parser():
    parser = optparse.OptionParser()
    parser.add_option('-r', '--results', action='store', type='string',
                      dest='results', default=DEFAULT_RESULTS_PATH,
                      help='Set path to results folder')
    parser.add_option('-m', '--show_mode', action='store', type='choice',
                      dest='show_mode', default='short',
                      choices=['short', 'medium', 'full'],
                      help='Set display mode information')
    parser.add_option('-s', '--start_date', action='callback', type='str',
                      callback=date_callback, dest='start_date',
                      help='Set start date of search')
    parser.add_option('-e', '--end_date', action='callback', type='str',
                      callback=date_callback, dest='end_date',
                      help='Set date of finishing search')
    parser.add_option('--vk_mode', action='store', type='choice',
                      dest='vk_mode', default=vk_mode_list[0],
                      choices=vk_mode_list,
                      help='Set options for searching VK. If set option ' \
                           '\'list\', the search will be made in users ' \
                           'VK-list (you must set --vk-list options)')
    parser.add_option('--vk_list', action='callback', callback=vararg_callback,
                      default=[], dest='vk_list',
                      help='Set list of VK\'s for search')
    parser.add_option('-p', '--run_program', action='callback',
                      callback=vararg_callback,
                      dest='run_program',
                      default=['utest', 'tst_run', 'tst_run1'],
                      help='Set running program for search')
    parser.add_option('--test_mode', action='store', type='choice',
                      dest='test_mode', default=test_mode_list[0],
                      choices=test_mode_list,
                      help='Set options for searching tests. If set option ' \
                           '\'list\', the search will be made in users ' \
                           'test-list (you must set --test-list options)')
    parser.add_option('--test_list', action='callback',
                      callback=vararg_callback, dest='test_list',
                      default=[], help='Set list of tests for search')
    parser.add_option('-v', '--vk_names', action='store_true', dest='vk_names',
                      help='Print vk and ip only')
    parser.add_option('-t', '--test_names', action='store_true',
                      dest='test_names',
                      help='Print test names and test types only')
    opts, args = parser.parse_args()
    return opts


def file_is_stp_results(res_file):
    # if argument is folder
    if not os.path.isfile(res_file):
        return False
    # name without path
    r_lim = res_file.rfind('/')
    name = res_file[r_lim + 1:]
    p = re.compile(
        '^[\d]{4}(0[1-9]|1[0-2])(0[1-9]|1[0-9]|2[0-9]|3[01])_[\d]{3}.res$')
    return not p.match(name) is None


def get_vk_list_from_res_yaml(res_yaml):
    for testing in res_yaml.values():
        for vk_name, vk in testing['vk'].items():
            yield (vk_name, vk.get('ip', 'unknown'))


def get_test_from_res_yaml(res_yaml):
    for testing in res_yaml.values():
        for vk in testing['vk'].values():
            for test_name, test in vk['tests'].items():
                yield (test_name, test.get('type', 'unknown'))


def check_vk(vk_name, vk, vk_mode, vk_list):
    # not set or 'All'
    if vk_mode is None or vk_mode == vk_mode_list[0]:
        return True
    # success
    elif vk_mode == vk_mode_list[1]:
        return vk['resume'] == 'success'
    # unsuccess
    elif vk_mode == vk_mode_list[2]:
        return vk['resume'] == 'unsuccess'
    # search by name
    elif vk_mode == vk_mode_list[3]:
        return vk_name in vk_list
    # search by ip
    elif vk_mode == vk_mode_list[4]:
        return vk.get('ip', 'unknown') in vk_list
    else:
        return False


def check_test(test_name, test, test_mode, test_list, run_prg_list=None):
    # run_program
    if run_prg_list == None or len(run_prg_list) == 0:
        return False
    elif not (test.get('test_program', 'utest') in run_prg_list):
        return False
    # not set or 'All'
    if test_mode is None or test_mode == test_mode_list[0]:
        return True
    # success
    elif test_mode == test_mode_list[1]:
        return test['errors'] == 0
    # unsuccess
    elif test_mode == test_mode_list[2]:
        return test['errors'] > 0
    # search by name
    elif test_mode == test_mode_list[3]:
        return test_name in test_list
    # search by type
    elif test_mode == test_mode_list[4]:
        return test['type'] in test_list
    else:
        return False


def testing_view_short(res_yaml, arg_opt):
    for testing in res_yaml.values():
        duration = str(datetime.timedelta(seconds=testing['duration']))
        print testing['date'].ljust(30) + duration.center(30) + testing['resume'].center(20).upper()


def testing_view_medium(res_yaml, arg_opt):
    for testing in res_yaml.values():
        testing_titles = list()
        duration = str(datetime.timedelta(seconds=testing['duration']))
        testing_titles.append(testing['date'].ljust(50) + duration.center(20) +\
                              testing['resume'].center(10).upper())
        testing_titles.append(testing['report'])
        testing_titles.append(80*'-')

        vk_lines = list()
        for vk_key, vk in testing['vk'].items():
            if check_vk(vk_key, vk, arg_opt.vk_mode, arg_opt.vk_list):
                duration = str(datetime.timedelta(seconds=vk['time']))
                vk_lines.append(vk_key.center(30) + \
                                vk.get('ip', 'unknown').ljust(20) + \
                                duration.center(20) + \
                                vk.get('resume').center(10))

        if len(vk_lines):
            for s in testing_titles + vk_lines:
                print s
            print ''


def testing_view_full(res_yaml, arg_opt):
    for testing in res_yaml.values():
        testing_lines = list()
        duration = str(datetime.timedelta(seconds=testing['duration']))
        testing_lines.append(testing['date'].ljust(40) + duration.center(20) + \
                             testing['resume'].center(20).upper())
        testing_lines.append(testing['report'])
        testing_lines.append(80 * '=')

        glb_vk_lines = list()
        for vk_key, vk in testing['vk'].items():
            vk_lines = list()
            if not check_vk(vk_key, vk, arg_opt.vk_mode, arg_opt.vk_list):
                continue
            duration = str(datetime.timedelta(seconds=vk['time']))
            vk_lines.append(vk_key.ljust(25) + \
                            vk.get('ip', 'unknown').center(15) + \
                            duration.center(20) + vk.get('resume').center(20))
            vk_lines.append('%s|%s|%s|%s' % ('test'.center(24),
                                             'type'.center(14),
                                             'errors'.center(19),
                                             'passes'.center(20)))
            vk_lines.append(80 * '-')

            test_lines = list()
            for test_key, test in vk['tests'].items():
                if not check_test(test_key, test, arg_opt.test_mode,
                                  arg_opt.test_list, arg_opt.run_program):
                    continue
                test_lines.append('%s|%s|%s|%s' % (test_key.center(24),
                                                      test['type'].center(14),
                                                      str(test['errors']).center(19),
                                                      str(test['passes']).center(20)))
            if len(test_lines):
                glb_vk_lines += vk_lines + test_lines

        if len(glb_vk_lines):
            for s in testing_lines + glb_vk_lines:
                print s
            print ''


def res_file_handler(res_file, arg_opt):
    start_date = arg_opt.start_date
    end_date = arg_opt.end_date

    rlim = res_file.rfind('/')
    llim = res_file.rfind('_')
    name = res_file[rlim + 1:llim]
    try:
        file_date = date_from_str(name, '%Y%m%d')
    except Exception:
        logging.warning('Results file (%s) has the wrong name' % (name))
        return

    if start_date:
        if file_date < start_date:
            return
    if end_date:
        if file_date > end_date:
            return
    try:
        res_yaml = yaml.load(file(res_file, 'r'))
    except yaml.YAMLError, err:
        if hasattr(err, 'problem_mark'):
            mark = err.problem_mark
            log.warning('File parsing errors. File: (%s) Error position: ' \
                        '(%s:%s)' % (res_file, mark.line + 1, mark.column + 1))
        else:
            log.warning('File parsing errors. File: (%s)' % res_file)
        return

    if arg_opt.vk_names:
        for n in get_vk_list_from_res_yaml(res_yaml):
            print n
    elif arg_opt.test_names:
        for t in get_test_from_res_yaml(res_yaml):
            print t
    elif arg_opt.show_mode == 'short':
        testing_view_short(res_yaml, arg_opt)
    elif arg_opt.show_mode == 'medium':
        testing_view_medium(res_yaml, arg_opt)
    elif arg_opt.show_mode == 'full':
        testing_view_full(res_yaml, arg_opt)


def search_in_folder(opt):
    try:
        dir_path = opt.results
        res_files = filter(lambda f: file_is_stp_results(f),
                           map(lambda x: '%s/%s' % (dir_path, x),
                               os.listdir(dir_path)))
    except OSError, err:
        logging.error(str(err) + '. Check path to results folder')
        return

    res_files.sort()
    for res_file in res_files:
        res_file_handler(res_file, opt)


def main():
    opt = opt_parser()
    search_in_folder(opt)


if __name__ == '__main__':
    main()
