#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Command line interface
"""

from cmd import Cmd
import os
import sys
import logging.config

import StpControl
# import Messages
from Common import APP_NAME, APP_VERSION


# -----------------------------------------------------------------------------
class CLI(Cmd):
    """Command line interface class"""

    # singleton
    def __new__(cls, *args, **kw):
        if not hasattr(cls, '_instance'):
            cls._instance = super(CLI, cls).__new__(cls, *args, **kw)
        return cls._instance

    def __init__(self):
        # cmd parser
        Cmd.__init__(self)
        self.prompt = "> "
        self.intro  = "%s version %s\n" \
                      "Copyright (C) 2015 MCST. All rights reserved.\n" \
                      "(type 'help' for getting help about commands)" \
                      % (APP_NAME, APP_VERSION)
        self.doc_header = "Available commands:\n" \
                          "(type 'help command' for info about command)"
        # Test Controller
        self.__stp = StpControl.StpControl()

    # ------------------------------------
    #           EXIT & etc.
    # ------------------------------------

    def do_EOF(self, args):
        """Exit application"""
        return self.do_exit(args)

    def do_quit(self, args):
        """Exit application"""
        return self.do_exit(args)

    def do_exit(self, args):
        """Exit application"""
        print("Terminate")
        return True

    def default(self, line):
        print ("Unknown command. For help type 'help'.")

    def emptyline(self):
        pass

    # ------------------------------------
    #           GLOBAL COMMANDS
    # ------------------------------------

    def do_state(self):
        """Print info about current status UI and STP"""
        print ("State info")

    # ------------------------------------
    #           NETWORK COMMANDS
    # ------------------------------------

    def do_connect(self, addr):
        """Connection to target host:port
        :param addr:
        """
        if not addr:
            self.__stp.stp_connect()
        elif ':' not in addr:
            self.__stp.stp_connect(addr)
        else:
            host, port = addr.split(':', 1)
            if not port.isdigit():
                print('Port must be digits.')
                logging.error('Port must be digits.')
                return
            self.__stp.stp_connect(host, port)

    def do_disconnect(self):
        """Disconnection from exist connection"""
        pass

    # ------------------------------------
    #           CFG COMMANDS
    # ------------------------------------

    def do_opencfg(self, args):
        """Read configuration from file: opencfg filename
        :param args:
        """
        if os.path.exists(args):
            if self.__stp.config_read(args):
                print("File [%s] open successfully" % args)
            else:
                print("Error: open file [%s]" % args)
        else:
            print ("Error: can't open file [%s]" % (args))

    def do_lscfg(self, args):
        """Print configuration
        :param args:
        """
        cfg = self.__stp.get_config()
        if cfg:
            print (cfg)
            cfg.dump()

    def do_newcfg(self, args):
        """Create empty cfg
        :param args:
        """
        self.__stp.config_create()

    def do_addvk(self, args):
        """Added vk to current cfg. if no cfg - created it
        :param args:
        """
        if not self.__stp.config_present():
            self.__stp.config_create()
        self.__stp.mdf_addvk()

    def do_delvk(self, args):
        """Delete vk from current cfg.
        :param args:
        """
        if not args:
            logging.warning("Not name for del vk")
            return

        if self.__stp.config_present():
            if self.__stp.mdf_delvk(args):
                logging.info("vk '%s' was deleted" % args)
            else:
                logging.error("Delete vk '%s' from cfg" % args)
        else:
            logging.warning("Attempt delete vk '%s' from empty cfg" % args)

    # ------------------------------------
    #           TEST COMMANDS
    # ------------------------------------

    def do_start(self, args):
        """Stop testing
        :param args:
        """
        self.__stp.test_start()

    def do_stop(self, args):
        """Stop testing
        :param args:
        """
        self.__stp.test_stop()

    # ------------------------------------
    def main(self):
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            logging.info("Quit CLI ...")
        sys.exit(0)


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    try:
        logging.config.fileConfig('log.conf')
    except IOError:
        print("Logging into file off")
    cli_instance = CLI()
    cli_instance.main()

