#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Configuration Model:

cfgroot
    hosts [{}]
        hostname
        options {}  # class: host_options
            concurrency
            enable
            errors
            ip
            max_sys_errors
            max_sys_passes
            max_sys_time
            passes
            program
            setup_program
            username
            realtime
            repeat
            singlepass
            time
        tests [{}]
            concurrency
            device
            enable
            instances [{}]
                args
                cpu
                errors
                max_sys_errors
                max_sys_passes
                max_sys_time
                param {}
                    algorithm
                    ip
                    mode
                    number
                    delay
                    port
                    size
                    time
                passes
                realtime
                repeat
                singlepass
                time
            program
            username
            realtime
            singlerun
            testname
            type
    links [{}]
        "host1"
             name
             test
        "host2"
             name
             test
    options {}
        autostart
        concurrency
        errors
        max_sys_errors
        max_sys_passes
        max_sys_time
        passes
        program
        setup_program
        username
        realtime
        repeat
        singlepass
        saveallresults
        reconnect_time
        time
    svg
"""

import os
import sys
import logging
import json
from binascii import crc32 as sum32
from copy import deepcopy
from subprocess import Popen, PIPE

from stp_gui import Common
from stp_gui.ExtraLogging import ExtraLog

def _(text):
    return Common._(text)
# ------------------------------------------------------------------------------
# GLOBAL PARAMETERS

REMDEF = True   # don't save key:data pair to .cfg if data empty


# ------------------------------------------------------------------------------
class ModelChange:
    MDF_CHN_VK = 1
    MDF_CHN_PREF = 2
    MDF_CHN_TEST = 3
    LOAD_CFG = 4
    NEW_CFG = 5

def strMode(mode):
    return {
        ModelChange.MDF_CHN_VK: 'change VK',
        ModelChange.MDF_CHN_PREF: 'change PREF',
        ModelChange.MDF_CHN_TEST: 'change TEST',
        ModelChange.LOAD_CFG: 'load cfg',
        ModelChange.NEW_CFG: 'new cfg',
    }.get(mode, 'unknown change')

# ------------------------------------------------------------------------------
class JSONSerializable:
    """
    Базовый класс для конфигурации

    В производных классах достаточно определить
        data_help = {...}  # справка по опциям
        data_def = {...}  # значения по-умолчанию
    и при необходимости перегрузить методы.

    data_help = {
        "<key>":  [<type>, <defval/classname>, "comment"],
        ...
        }

        <key>:
          list - for []
          dict - for {}
          bool - for 0,1
          int  - for 0,>0
          str  - for ""

        <defval/classname>:
          class name for list, dict
          default value for bool, int, str
    """

    def __init__(self, json_data=None):
        self.parent = None
        self.root = None
        self.data = dict()
        if json_data:
            self.load(json_data)
        else:
            self.data = deepcopy(self.data_def)

    # -------------------------------------------------
    @staticmethod
    def _make_list(name, json_list):
        """Make list/array and add all items"""
        lst = []
        kls = globals()[name]
        if kls:
            for i in range(len(json_list)):
                obj = kls(json_list[i])
                if obj:
                    lst.append(obj)
        return lst

    @staticmethod
    def _make_dict(name, json_dict):
        """Make dict and add all items"""
        kls = globals()[name]
        if kls:
            return kls(json_dict)
        else:
            return None

    # -------------------------------------------------
    # TODO: create setters for data
    # @property
    # def data(self):
    #    return self.__data
    # @data.setter
    # def data(self, value):
    #    if self.__data != data:
    #        self.__data = data
    #        self.changed()

    def changed(self):
        if self.root:
            self.root.notify_observers()

    # -------------------------------------------------
    def load(self, d):
        """Load from dictionary
        :param d:
        """
        for key in self.data_def.keys():
            if key in d:
                if self.data_help[key][0] == list:
                    self.data[key] = \
                        JSONSerializable._make_list(self.data_help[key][1],
                                                    d.pop(key))
                elif self.data_help[key][0] == dict:
                    self.data[key] = \
                        JSONSerializable._make_dict(self.data_help[key][1],
                                                    d.pop(key))
                else:  # bool, int, str
                    self.data[key] = d.get(key)
            else:
                self.data[key] = deepcopy(self.data_def[key])

    def save(self, remdef=True):
        """Dump to json
        :param remdef:
        """
        data = deepcopy(self.data)
        # remove pair if data empty
        if remdef:
            for key in self.data.keys():
                if data[key] == self.data_def[key]:
                    data.pop(key)
        return data

    def dump(self, compact=False, remdef=False):
        """dump classes tree to json
        :param compact:
        """
        if compact:
            return json.dumps(self.save(remdef),
                              cls=AutoJSONEncoder,
                              sort_keys=True,
                              ensure_ascii=False,
                              separators=(',', ':'))
        else:
            return json.dumps(self.save(remdef),
                              cls=AutoJSONEncoder,
                              sort_keys=True,
                              indent=4,
                              ensure_ascii=False,
                              separators=(',', ': '))

    def dumpchild(self, key=None, compact=False):
        """dump child classes tree to json
        :param key:
        :param compact:
        """
        child = None
        for k in self.data.keys():
            if k == key:
                child = self.data[k]
                break
        if not child:
            return
        if compact:
            return json.dumps(child,
                              cls=AutoJSONEncoder,
                              sort_keys=True,
                              ensure_ascii=False,
                              separators=(',', ':'))
        else:
            return json.dumps(child,
                              cls=AutoJSONEncoder,
                              sort_keys=True,
                              indent=4,
                              ensure_ascii=False,
                              separators=(',', ': '))

    def __repr__(self):
        """Return string representation of a dictionary"""
        return '%s' % self.data

    __str__ = __repr__

    def __getattr__(self, key):
        """for cycle iterator"""
        if key == 'data':
            raise AttributeError(KeyError)
        try:
            return self.data[key]
        except KeyError as k:
            raise AttributeError(k)

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError as k:
            raise AttributeError(k)


# -----------------------------------------------------------------------------
class param(JSONSerializable):
    """Словарь param - аргументы utest"""

    data_help = {}
    data_def = {}

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError as k:
            raise AttributeError(k)

# -----------------------------------------------------------------------------
class instance(JSONSerializable):
    """Словарь instance - описание экземпляра теста для ВК"""

    data_help = {}
    data_def = {}

    def save(self, remdef=True):
        data = deepcopy(self.data)

        if remdef:
            param = data['param']
            flag = False
            for key in param.data.keys():
                if param.data[key] != param.data_def[key]:
                    flag = True
                    break

            if not flag:
                data.pop('param')

            rmkeys = []
            for key in data.keys():
                if data[key] == self.data_def[key]:
                    rmkeys.append(key)

            for key in rmkeys:
                data.pop(key)


        return data


# -----------------------------------------------------------------------------
class test(JSONSerializable):
    """Словарь test - описание теста для ВК"""

    data_help = {}
    data_def = {}

    def save(self, remdef=True):
        data = deepcopy(self.data)
        if remdef and len(data['instances']) == 1:
            inst = data['instances'][0]
            if inst.save() == {}:
                data.pop('instances')
        rmkeys = []
        for key in data:
            if data[key] == self.data_def[key]:
                rmkeys.append(key)

        for key in rmkeys:
            data.pop(key)

        return data

    def __gt__(self, test2):
        return self.testname > test2.testname

# -----------------------------------------------------------------------------
class host_options(JSONSerializable):
    """Словарь options - настройки тестирования ВК"""

    data_help = {}
    data_def = {}


# -----------------------------------------------------------------------------
class host(JSONSerializable):
    """Словарь host - тестируемый ВК"""

    data_help = {
        "hostname": [str, "vk", "Имя ВК"],
        "options": [dict, "host_options", "Настройки тестирования ВК"],
        "tests": [list, "test", "Список тестов для ВК"],
    }
    data_def = {
        "hostname": "vk",
        "options": host_options(),
        "tests": [],
    }

    def save(self, remdef=True):
        """Dump to json
        :param remdef:
        """
        data = deepcopy(self.data)
        if remdef:
            opts = data['options']
            if opts.save() == {}:
                data.pop('options')

            if data['tests'] == []:
                data.pop('tests')
        return data

    @property
    def hostname(self):
        return self.data['hostname']

    @hostname.setter
    def hostname(self, value):
        self.data['hostname'] = value

    def __gt__(self, host2):
        return self.hostname > host2.hostname

# -----------------------------------------------------------------------------
class pair(JSONSerializable):
    """Словарь pair - описания одного ВК для связи"""

    data_help = {
        "name": [str, "", "Имя ВК"],
        "test": [str, "", "Имя теста"],
    }
    data_def = {
        "name": "",
        "test": ""
    }


# -----------------------------------------------------------------------------
class link(JSONSerializable):
    """
    Словарь link - описание связи между тестами
        словарь из 2-х элементов:
        {"host1" :{"name":"имя_ВК_1", "test":"имя_теста_1"},
        {"host2" :{"name":"имя_ВК_2", "test":"имя_теста_2"}
    """

    data_help = {
        "host1": [dict, "pair", "Имя первого ВК"],
        "host2": [dict, "pair", "Имя второго ВК"],
    }
    data_def = {
        "host1": {},
        "host2": {}
    }


# -----------------------------------------------------------------------------
class options(JSONSerializable):
    """Словарь options - общие настройки тестирования"""

    data_help = {}
    data_def = {}

    def save(self, remdef=True):
        """Dump to json all root options - overload
        :param remdef:
        """
        empty_str_defs = {k: v for k, v in self.data_def.items() if v == ''}
        new_data = deepcopy(self.data)
        for k in empty_str_defs.keys():
            if new_data[k] == '':
                new_data.pop(k)
        return new_data


# -----------------------------------------------------------------------------
class cfgroot(JSONSerializable):
    """Словарь верхнего уровня"""

    data_help = {
        "hosts": [list, "host", "Перечисление тестируемых ВК"],
        "links": [list, "link", "Описание связей между тестами"],
        "options": [dict, "options", "Общие настройки тестирования"],
        "svg": [str, "svg", "SVG представление конфигурации"],
    }
    data_def = {
        "hosts": [],
        "links": [],
        "options": options(),
        "svg": "",
    }


# -----------------------------------------------------------------------------
class AutoJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        """method to serialize my types to json (for json.dump)
        :param obj:
        """
        try:
            # run save() from my classes
            return obj.save(remdef=REMDEF)
        except AttributeError:
            return json.JSONEncoder.default(self, obj)


# -----------------------------------------------------------------------------
class Config(ExtraLog):
    """File operations for Config"""

    closed = False

    # singleton
    def __new__(cls, *args, **kw):
        if not hasattr(cls, '_instance'):
            cls._instance = super(Config, cls).__new__(cls, *args, **kw)
            cls._instance.__initialized = False
        return cls._instance

    def __init__(self):
        if self.__initialized:
            # print 'in init'
            return
        self.__cfgfilename = None  # current cfg filename
        self.__cfgchanged = False  # current cfg state (True == changed)
        self.__cfgroot = None
        self.__cfgschemaparser = SchemaParser()
        self.__cfgschemaparser.loadSchema('tune/stp_cfg.schema')
        self.__cfgschemaparser.createProperties()

        self._observers = []  # наблюдатели

        self.__initialized = True

    # -------------------------------------------------
    def add_observer(self, inobserver):
        self._observers.append(inobserver)

    def rem_observer(self, inobserver):
        self._observers.remove(inobserver)

    def notify_observers(self, mode, dump):
        self.debug(strMode(mode))
        if mode == ModelChange.MDF_CHN_VK or \
                mode == ModelChange.MDF_CHN_PREF or \
                mode == ModelChange.MDF_CHN_TEST:
            self.__cfgchanged = True
        for x in self._observers:
            x.model_changed()

    # -------------------------------------------------

    def get_filename(self):
        """return current config filename"""
        return self.__cfgfilename

    def is_changed(self):
        """return config state"""
        return self.__cfgchanged

    def is_empty(self):
        """return true if config empty"""
        if self.__cfgroot == None:
            return True
        else:
            return False

    def get_cfg(self):
        """return config dictionary"""
        return self.__cfgroot

    def get_hosts(self):
        """return hosts dictionary"""
        return self.__cfgroot.hosts

    def set_hosts(self, value):
        self.cfg.data['hosts'] = value

    def get_options(self):
        """return options dictionary"""
        return self.__cfgroot.options

    def set_options(self, value):
        self.cfg.data['options'] = value

    def get_links(self):
        """return config dictionary"""
        return self.__cfgroot.links

    def set_links(self, value):
        self.cfg.data['liks'] = value

    def get_svg(self):
        """return svg string"""
        return self.__cfgroot.svg

    def set_svg(self, value):
        self.cfg.data['svg'] = value

    def get_schema_parser(self):
        """return config schema"""
        return self.__cfgschemaparser

    # TODO: convert to decorator
    cfg = property(get_cfg, None, "Config")
    hosts = property(get_hosts, set_hosts, None, "Config")
    options = property(get_options, set_options, None, "Config")
    links = property(get_links, set_links, None, "Config")
    svg = property(get_svg, set_svg, None, "Config")

    def create(self, jsonstr=None):
        """create empty config
        :param jsonstr:
        """
        self.__cfgfilename = None
        self.__cfgchanged = False
        if jsonstr:
            self.__cfgroot = cfgroot(json.loads(jsonstr))
        else:
            self.__cfgroot = cfgroot()
        self.notify_observers(ModelChange.NEW_CFG, self.__cfgroot.dump(compact=True))
        return self.__cfgroot

    def read(self, filename):
        """read config from file
        :param filename: full filename
        """
        try:
            f = open(str(filename), 'r')
            f_str = f.read()
            if f_str[0:3] == "STP":
                prog = os.path.join(Common.bin_dir, 'converter')
                if os.path.isfile(prog):
                    process = Popen([prog, filename], stdout=PIPE)
                    (output, err) = process.communicate()
                    exit_code = process.wait()
                    if exit_code:
                        # TODO Different messages based on exit code
                        pass
                    self.__cfgroot = cfgroot(json.loads(output))
                else:
                    logging.warning(_("Converter must be compiled to work with simple config"))
                    self.__cfgroot = cfgroot(json.loads("ERROR"))
            else:
                self.__cfgroot = cfgroot(json.loads(f_str))
            f.close()
        except IOError:
            return False
        self.__cfgfilename = filename
        self.__cfgchanged = False
        self.notify_observers(ModelChange.LOAD_CFG, self.dump(compact=True))
        return True

    def prepare_to_dump(self):
        self.hosts.sort()
        for h in self.hosts:
            h.tests.sort()

    def dump(self, compact=False, remdef=False):
        if not self.__cfgroot:
            return None
        self.prepare_to_dump()
        """dump classes tree to json
        :param compact: True for network, False for file
        """
        if compact:
            return json.dumps(self.__cfgroot.save(remdef),
                              cls=AutoJSONEncoder,
                              sort_keys=True,
                              ensure_ascii=False,
                              separators=(',', ':'))
        else:
            return json.dumps(self.__cfgroot.save(remdef),
                              cls=AutoJSONEncoder,
                              sort_keys=True,
                              indent=4,
                              ensure_ascii=False,
                              separators=(',', ': '))

    def write(self, filename=None):
        """save config to file
        :param filename: full filename
        """
        if not filename:
            filename = self.__cfgfilename
        try:
            data = self.dump()
            # backup original file
            if os.path.exists(str(filename)):
                os.rename(str(filename), "%s.bak" % (filename))
            # save file
            try:
                f = open(str(filename), 'w')
                try:
                    f.write(data)
                except (OSError, IOError) as e:
                    logging.error(_('Error while saving file') + ': ' + e.strerror)
                    f.close()
                    return False
                f.close()
            except (OSError, IOError) as e:
                logging.error(_('Error while saving file') + ': ' + e.strerror)
                return False
        except (OSError, IOError):
            return False
        self.__cfgfilename = filename
        self.__cfgchanged = False
        return True

    def crc(self):
        if not self.__cfgroot:
            return 0

        json = self.dump(compact=True, remdef=True)
        # json = json.encode(encoding='UTF-8')
        return sum32(json.encode('utf-8'))

# -----------------------------------------------------------------------------
class SchemaParser():
    def __init__(self):
        self.schema = None

    def getDataHelp(self, options):
        data_help = {}
        for opt in options.items():
            if opt[1]['type'] == 'array':
                data_help[opt[0]] = [list, 'instance', opt[1]['description']]
            elif opt[1]['type'] == 'integer':
                if opt[1].get('enum') == [0, 1]:
                    data_help[opt[0]] = [bool, opt[1]['default'], opt[1]['description']]
                else:
                    data_help[opt[0]] = [int, opt[1]['default'], opt[1]['description']]

            elif opt[1]['type'] == 'string':
                if opt[1]['default'] is None:
                    data_help[opt[0]] = [str, '', opt[1]['description']]
                else:
                    data_help[opt[0]] = [str, opt[1]['default'], opt[1]['description']]
            elif opt[1]['type'] == 'object':
                data_help[opt[0]] = [dict, opt[0], opt[1]['description']]
            else:
                data_help[opt[0]] = [opt[1]['type'], opt[1]['default'], opt[1]['description']]
        return data_help

    def getDataDef(self, options):
        data_def = {}
        for opt in options.items():
            # print 'opt = ', opt
            if opt[1]['type'] == 'array' and opt[0] == 'instances':
                data_def[opt[0]] = [instance()]
            elif opt[1]['type'] == 'array':
                data_def[opt[0]] = []
            elif opt[1]['type'] == 'string':
                data_def[opt[0]] = ''
            elif opt[1]['type'] == 'object' and opt[0] == 'param':
                data_def[opt[0]] = param()
            else:
                data_def[opt[0]] = opt[1]['default']
        return data_def

    def loadSchema(self, filename):
        try:
            schema_path = os.path.join(Common.stp_dir, filename)
            with open(schema_path) as f:
                self.schema = json.load(f)
        except IOError as e:
            message = "Error while loading schema: " + e.strerror + ':' + \
                      e.filename
            logging.critical(message)
            sys.exit(message)
        # TODO: added catch for exception

        self.paramOptions = \
            self.schema['properties']['hosts']['items'][0]['properties']['tests']['items']['properties']['instances'][
                'items']['properties']['param']['properties']
        param.data_help = self.getDataHelp(self.paramOptions)
        param.data_def = self.getDataDef(self.paramOptions)

        self.instanceOptions = \
            self.schema['properties']['hosts']['items'][0]['properties']['tests']['items']['properties']['instances'][
                'items']['properties']
        instance.data_help = self.getDataHelp(self.instanceOptions)
        instance.data_def = self.getDataDef(self.instanceOptions)

        self.testOptions = self.schema['properties']['hosts']['items'][0]['properties']['tests']['items']['properties']
        # print 'testOptions: ', self.testOptions['instances']
        test.data_help = self.getDataHelp(self.testOptions)
        test.data_def = self.getDataDef(self.testOptions)

        self.hostOptions = self.schema['properties']['hosts']['items'][0]['properties']['options']['properties']
        host_options.data_help = self.getDataHelp(self.hostOptions)
        host_options.data_def = self.getDataDef(self.hostOptions)
        host.data_def['options'] = host_options()

        self.rootProperties = self.schema['properties']['options']['properties']
        options.data_help = self.getDataHelp(self.rootProperties)
        options.data_def = self.getDataDef(self.rootProperties)
        # Update cfgroot class, since it was initialized with old options()
        cfgroot.data_def['options'] = options()

    def createProp(self, target_class, options, mode):
        for opt in options.keys():
            def make_f(opt):
                def f_set(obj, val):
                    if obj.data[opt] == val:
                        return
                    obj.data[opt] = val

                def f_get(obj):
                    return obj.data[opt]

                return f_get, f_set

            funcs = make_f(opt)
            setattr(target_class, opt, property(funcs[0], funcs[1]))

    def createProperties(self):
        self.createProp(host_options, self.hostOptions, ModelChange.MDF_CHN_VK)
        self.createProp(options, self.rootProperties, ModelChange.MDF_CHN_PREF)
        self.createProp(test, self.testOptions, ModelChange.MDF_CHN_TEST)
        self.createProp(instance, self.instanceOptions, ModelChange.MDF_CHN_TEST)
        self.createProp(param, self.paramOptions, ModelChange.MDF_CHN_TEST)

# -----------------------------------------------------------------------------
# No __main__
