#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains customized Dialogs which are used by STP GUI
"""

from PyQt5 import QtGui, QtCore
from PyQt5.QtWidgets import *
import stp_gui.Common


def _(text):
    return stp_gui.Common._(text)


class RadioButtonDialog(QDialog):
    """
    Base class for dialogs using radio buttons.
    """
    def __init__(self, items_list):

        super(RadioButtonDialog, self).__init__()
        self.items = items_list
        items = map(_, self.items)

        self.buttons = []
        for item in items:
            self.buttons.append(QRadioButton(item))

        self.buttons[0].setChecked(True)

        v_layout = QVBoxLayout()

        for button in self.buttons:
            v_layout.addWidget(button)


        buttons = QDialogButtonBox()
        buttons.setStandardButtons(QDialogButtonBox.Ok
                                   | QDialogButtonBox.Cancel)
        buttons.button(QDialogButtonBox.Ok).setText(_('OK'))
        buttons.button(QDialogButtonBox.Cancel).setText(_('Cancel'))
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        v_layout.addWidget(buttons)
        self.setLayout(v_layout)

    def get_item(self):
        """
        Returns item selected by user
        """
        for (index, button) in enumerate(self.buttons):
            if button.isChecked():
                break
        item = self.items[index]
        return item


class CloseDialog(RadioButtonDialog):
    """
    Dialog displayed on exit from STP
    """

    def __init__(self):
        items = ['Terminate GUI and kernel', 'Terminate GUI only',
                 'Terminate all']
        super(CloseDialog, self).__init__(items)
        self.setWindowTitle(_('Quit?'))

        settings = QtCore.QSettings()
        last_value = str(settings.value('last_session/exit_type'))
        index = 0
        try:
            index = self.items.index(last_value)
        except ValueError:
            pass

        self.buttons[index].setChecked(True)


class TerminationDialog(RadioButtonDialog):
    """
    Dialog to choose which parts of STP should be terminated
    """

    def __init__(self):
        items = ['Terminate kernel only', 'Terminate all']
        super(TerminationDialog, self).__init__(items)
        self.setWindowTitle(_('Terminate...'))


class ConnectDialog(QInputDialog):
    """
    Dialog displayed when connecting to kernel
    """
    def __init__(self, items_list):

        super(ConnectDialog, self).__init__()
        self.setComboBoxItems(items_list)
        self.setComboBoxEditable(True)
        self.setWindowTitle(_('Connect to...'))
        self.setLabelText(_('Hostname or IP:'))
        self.setCancelButtonText(_('Cancel'))
        self.setOkButtonText(_('OK'))

    def exec_(self):
        """
        Show dialog and return values entered by user
        """
        super(ConnectDialog, self).exec_()
        return self.textValue(), self.result()


class StyleDialog(QInputDialog):
    """
    Dialog displayed when select style for window
    """
    def __init__(self, items_list):

        super(StyleDialog, self).__init__()
        self.setComboBoxItems(items_list)
        self.setComboBoxEditable(True)
        self.setWindowTitle(_('Choose style'))
        self.setLabelText(_('Available styles') + ':')
        self.setCancelButtonText(_('Cancel'))
        self.setOkButtonText(_('OK'))

    def exec_(self):
        """
        Show dialog and return values entered by user
        """
        super(StyleDialog, self).exec_()
        return self.textValue(), self.result()


class SelectConfigDialog(QMessageBox):
    """
    Dialog displayed when user has to choose between
    local config and one received from kernel
    """
    def __init__(self):
        super(SelectConfigDialog, self).__init__()

        self.setWindowTitle(_('Config mismatch'))
        text = "Local config is different from kernel's. "\
        "Would you like to use remote config?"
        self.setText(_(text))
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        self.button(QMessageBox.Yes).setText(_('Use remote config'))
        self.button(QMessageBox.No).setText(_('Use local config'))


class AddTestDialog(QDialog):
    """
    Dialog for adding new test. Allows to specify name and type
    """
    def __init__(self):
        super(AddTestDialog, self).__init__()

        grid = QGridLayout()

        self.line = QLineEdit()
        self.combo = QComboBox()
        self.combo.setEditable(True)
        name_label = QLabel(_('Name'))
        type_label = QLabel(_('Type'))

        button_box = QDialogButtonBox(QDialogButtonBox.Ok |
                                            QDialogButtonBox.Cancel)

        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        grid.addWidget(type_label, 0, 0)
        grid.addWidget(name_label, 0, 1)
        grid.addWidget(self.combo, 1, 0)
        grid.addWidget(self.line, 1, 1)
        grid.addWidget(button_box, 2, 1)

        self.combo.editTextChanged.connect(self.line.setText)

        self.setLayout(grid)

    def get_result(self):
        """
        Returns item selected by user
        """
        return self.line.text(), self.combo.currentText()

    def set_combo_items(self, items):
        """
        Fill combobox with provided values
        """
        self.combo.addItems(items)


class SimpleConfigDialog(QMessageBox):
    """
    Dialog which appears when user wants to save file loaded from simple config
    """
    def __init__(self):
        super(SimpleConfigDialog, self).__init__()

        self.setWindowTitle(_('Overwrite file?'))
        text = _('Are you sure you want to overwrite file with simple configuration?')
        self.setText(text)
        self.setStandardButtons(QMessageBox.Yes |
                QMessageBox.Save | QMessageBox.Cancel)
        self.button(QMessageBox.Save).setText(_('Save as...'))
