#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Qt GUI Viewer

 ----------------
| windowtitle [X]|
+----------------+
|       |        |
|  cfg  |  opt   |
|  view |  view  |
|       |        |
+----------------+
|    loglist     |
+----------------+
|   statusbar    |
 ----------------
"""

import json
import locale
import logging
import os
import subprocess
import sys
import time
from copy import deepcopy

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QFont, QFontMetrics

import stp_gui.Common as Common

try:
    from PyQt5 import QtWebKit
    from stp_gui.Monitoring import MonitoringWindow

    WEBKIT_INSTALLED = True
except ImportError:
    WEBKIT_INSTALLED = False

from PyQt5 import QtNetwork
from stp_gui.StpControl import StpControl
from stp_gui.ExtraLogging import ExtraLog
from stp_gui.StpNetwork import NetCmd
from stp_gui.Config import Config, host, test, options, instance
from stp_gui.Widgets import PropertiesWidget, TreeWidget, ResultsWidget, LogViewer, statusbar
from stp_gui.Dialogs import (TerminationDialog, CloseDialog, ConnectDialog, StyleDialog, SimpleConfigDialog,
                                   SelectConfigDialog)
from stp_gui.Settings import SettingsWindow

DEBUG = False

def _(text):
    return Common._(text)


class logHandler(logging.Handler):

    def setLogWidget(self, widget):
        self.widget = widget

    def emit(self, record):
        self.widget.append((time.strftime('%X - %d.%m'), record.levelname, record.message))

# ------------------------------------------------------------------------------
class windowtitle():
    """Set Window Title: <APP_NAME> - status"""

    def __init__(self):
        #self.__window = window
        self.__msg = ''

    def windowtitle_update_run(self, msg=None):
        """
        Update 'Test Run' flag

        @param msg: show info in title
        @type run: string

        @raise: no special exceptions

        @return: no return
        @rtupe: None
        """
        if msg is not None:
            self.__msg = msg
        self.windowtitle_update()

    def windowtitle_update(self):
        """Update message in window title"""
        s = Common.APP_NAME + ' - ' + self.__msg
        #self.__window.setWindowTitle(s)

#------------------------------------------------------------------------------
class GUIQT(ExtraLog, QtWidgets.QMainWindow, windowtitle):
    """This is an stp_gui application"""

    # singleton
    def __new__(cls, *args, **kw):
        if not hasattr(cls, '_instance'):
            cls._instance = super(GUIQT, cls).__new__(cls, *args, **kw)
            cls._instance.__initialized = False
        return cls._instance

    # ------------------------------------
    # CREATE GUI
    # ------------------------------------

    def initMenu(self):
        """Create Main Menu"""
        menu = QtWidgets.QMenuBar()
        fileMenu = menu.addMenu(_("&Config"))

        self.newFileAction = fileMenu.addAction(_("&New"))
        self.newFileAction.setShortcut('Ctrl+N')
        self.newFileAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_FileIcon))
        self.newFileAction.triggered.connect(self.newFile)
        self.openAction = fileMenu.addAction(_("&Open"))
        self.openAction.setShortcut('Ctrl+O')
        self.openAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_DirOpenIcon))
        self.openAction.triggered.connect(lambda: self.openFile())

        self.recentMenu = fileMenu.addMenu(_("Open &Recent"))
        self.updateRecentMenu()
        closeAction = fileMenu.addAction(_("Close"))
        closeAction.triggered.connect(self.closeFile)

        fileMenu.addSeparator()

        embedSVGAction = fileMenu.addAction(_("Embed SVG..."))

        def embed_svg_handler():
            cfg = Config()
            if cfg.is_empty():
                logging.info(_("Embedding SVG - ignored. No config loaded"))
                return
            if cfg.svg:
                button = QtWidgets.QMessageBox.question(
                    self, _('Replace SVG?'),
                    _('Config already has svg. Do you want to replace it?'),
                    QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.Cancel)
                if button == QtWidgets.QMessageBox.Cancel:
                    return

            filename = QtWidgets.QFileDialog.getOpenFileName(
                self, _('Open file'),
                '', _('SVG graphic Files (*.svg);;All Files (*)'), None,
                QtWidgets.QFileDialog.DontUseNativeDialog)
            if not os.path.exists(str(filename)):
                return
            else:
                with open(str(filename), 'r') as svg_file:
                    svg = svg_file.read()
                    svg_str = svg.replace('\n', '\\n').replace('"', '\\"')
                    cfg.svg = svg_str
                    logging.info(_('SVG added successfully'))
                    self.stpCtrl.send_cfg()

        embedSVGAction.triggered.connect(embed_svg_handler)

        deleteSVGAction = fileMenu.addAction(_("Delete SVG"))

        def del_svg_handler():
            cfg = Config()
            if cfg.is_empty():
                logging.info(_("Deleting SVG - ignored. No config loaded"))
                return
            if not cfg.svg:
                logging.info(_("Deleting SVG - ignored. No SVG in config"))
                return

            cfg.svg = ""
            logging.info(_('SVG deleted successfully'))
            self.monitoringWin.closeMonitoring()
            self.monitoringWin = None
            self.stpCtrl.send_cfg()

        deleteSVGAction.triggered.connect(del_svg_handler)

        fileMenu.addSeparator()

        self.saveAction = fileMenu.addAction(_("Save"))
        self.saveAction.setShortcut('Ctrl+S')
        self.saveAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_DialogSaveButton))
        self.saveAction.triggered.connect(self.saveFile)

        saveAsAction = fileMenu.addAction(_("Save As.."))
        saveAsAction.triggered.connect(self.saveFileAs)

        fileMenu.addSeparator()

        self.quitAction = fileMenu.addAction(_("&Quit"))
        self.quitAction.setShortcut('Ctrl+Q')
        self.quitAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_DialogCloseButton))
        self.quitAction.triggered.connect(self.close)

        def svg_enable():
            cfg = Config()
            if cfg.is_empty():
                deleteSVGAction.setEnabled(False)
                embedSVGAction.setEnabled(False)
            elif cfg.svg:
                deleteSVGAction.setEnabled(True)
                embedSVGAction.setEnabled(True)
            else:
                deleteSVGAction.setEnabled(False)
                embedSVGAction.setEnabled(True)

        fileMenu.aboutToShow.connect(svg_enable)

        self.editMenu = menu.addMenu(_("&Edit"))

        self.copyAction = self.editMenu.addAction(_("Copy"))
        self.copyAction.setShortcut('Ctrl+C')
        self.copyAction.triggered.connect(self.cfgTree.copyNode)

        self.pasteAction = self.editMenu.addAction(_("Paste"))
        self.pasteAction.setShortcut('Ctrl+V')
        self.pasteAction.triggered.connect(self.cfgTree.pasteNode)
        self.editMenu.addSeparator()

        self.addVK = self.editMenu.addAction(_("Add VK"))
        self.addVK.triggered.connect(self.cfgTree.addVK)

        self.addTest = self.editMenu.addAction(_("Add test"))
        self.addTest.triggered.connect(self.cfgTree.addTest)

        self.editMenu.addSeparator()

        self.delVK = self.editMenu.addAction(_("Delete VK"))
        self.delVK.triggered.connect(self.cfgTree.delVK)

        self.delTest = self.editMenu.addAction(_("Delete test"))
        self.delTest.triggered.connect(self.cfgTree.delTest)

        self.editMenu.addSeparator()

        self.renameAction = self.editMenu.addAction(_("Rename"))
        self.renameAction.setShortcut('F2')
        self.renameAction.triggered.connect(self.cfgTree.renameNode)

        self.editMenu.addSeparator()

        '''
        undoAction = editMenu.addAction(_("Undo"))
        undoAction.triggered.connect(self.undo)

        redoAction = editMenu.addAction(_("Redo"))
        redoAction.triggered.connect(self.redo)
        '''

        viewMenu = menu.addMenu(_("&View"))

        # TODO add implementation for marking all tests enabled/disabled
        '''
        selectAllAction = viewMenu.addAction(_("Select all"))
        selectAllAction.triggered.connect(self.selectAll)

        deselectAllAction = viewMenu.addAction(_("Deselect all"))
        deselectAllAction.triggered.connect(self.deselectAll)
        '''

        # toggleNodeAction = viewMenu.addAction(_("Display hidden"))
        # toggleNodeAction.triggered.connect(self.toggleNode)

        expandConfigTree = viewMenu.addAction(_("Expand tests level"))

        def expand():
            iterator = QtWidgets.QTreeWidgetItemIterator(self.cfgTree)
            while iterator.value():
                item = iterator.value()
                item.setExpanded(True)
                iterator += 1

        expandConfigTree.triggered.connect(expand)

        collapseConfigTree = viewMenu.addAction(_("Expand hosts level"))

        def collapse():
            iterator = QtWidgets.QTreeWidgetItemIterator(self.cfgTree)
            while iterator.value():
                item = iterator.value()
                item.setExpanded(False)
                iterator += 1
            self.cfgTree.getNodeByPath('/')

        collapseConfigTree.triggered.connect(collapse)

        viewMenu.addSeparator()

        styleAction = viewMenu.addAction(_('Window style'))
        styleAction.triggered.connect(self.chooseStyle)

        viewMenu.addSeparator()

        # maxTreeAction = viewMenu.addAction(_("Maximize Tree"))
        # maxTreeAction.triggered.connect(self.maxTree)

        # maxCfgAction = viewMenu.addAction(_("Maximize Options"))
        # maxCfgAction.triggered.connect(self.maxCfg)

        # viewMenu.addSeparator()

        self.monitoringWin = None
        self.hotkeys_win = None
        showMonitoring = viewMenu.addAction(_("Show monitoring window"))
        showMonitoring.setEnabled(WEBKIT_INSTALLED)
        showMonitoring.triggered.connect(self.showMonitoring)

        testsMenu = menu.addMenu(_("&Testing"))
        testsMenu.addSeparator()

        self.runTestsAction = testsMenu.addAction(_("&Run Tests"))
        self.runTestsAction.setShortcut('F5')
        self.runTestsAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_MediaPlay))
        self.runTestsAction.triggered.connect(self.runTests)

        self.stopTestsAction = testsMenu.addAction(_("&Stop Tests"))
        self.stopTestsAction.setShortcut('F6')
        self.stopTestsAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_MediaStop))
        self.stopTestsAction.triggered.connect(self.stopTests)

        self.pauseTestsAction = testsMenu.addAction(_("&Pause Tests"))
        self.pauseTestsAction.setShortcut('F5')
        self.pauseTestsAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_MediaPause))
        self.pauseTestsAction.triggered.connect(self.pauseTests)

        testsMenu.addSeparator()

        self.resetTestsAction = testsMenu.addAction(_("Rese&t Tests"))
        self.resetTestsAction.setShortcut('F7')
        self.resetTestsAction.setIcon(
            QtWidgets.qApp.style().standardIcon(
                QtWidgets.QStyle.SP_DialogResetButton))
        self.resetTestsAction.triggered.connect(self.resetTests)

        settingsMenu = menu.addMenu(_("&Control"))

        self.connectTestsAction = settingsMenu.addAction(_("&Connect to kernel..."))
        self.connectTestsAction.triggered.connect(self.netConnect)
        if os.path.exists(Common.data_dir + '/connected.png'):
            self.connectTestsAction.setIcon(QIcon(Common.data_dir + '/connected.png'))
        else:
            self.connectTestsAction.setIcon(
                QtWidgets.qApp.style().standardIcon(
                    QtWidgets.QStyle.SP_ComputerIcon))
        self.connectTestsAction.setShortcut('F8')

        self.startKernelAction = settingsMenu.addAction(_("Start kernel"))
        self.startKernelAction.triggered.connect(self.startKernel)

        self.disconnectTestsAction = settingsMenu.addAction(_("&Disconnect from kernel"))
        self.disconnectTestsAction.triggered.connect(self.netDisconnect)
        self.disconnectTestsAction.setVisible(False)
        if os.path.exists(Common.data_dir + '/disconnected.png'):
            self.disconnectTestsAction.setIcon(QIcon(Common.data_dir + '/disconnected.png'))
        else:
            self.disconnectTestsAction.setIcon(
                QtWidgets.qApp.style().standardIcon(
                    QtWidgets.QStyle.SP_DialogCloseButton))

        self.disconnectTestsAction.setShortcut('F8')

        self.reconnectAction = settingsMenu.addAction(_("&Reconnect VK"))
        self.reconnectAction.setShortcut('Ctrl+R')
        self.reconnectAction.triggered.connect(self.netReconnect)

        self.terminateAction = settingsMenu.addAction(_("&Terminate"))
        self.terminateAction.triggered.connect(self.showTerminateDialog)

        self.exclusiveModeAction = settingsMenu.addAction(_("&Exclusive mode"))
        self.exclusiveModeAction.setShortcut('Ctrl+E')
        self.exclusiveModeAction.triggered.connect(self.exclusiveMode)

        # showSettings = settingsMenu.addAction(_("Show gui settings"))
        # showSettings.triggered.connect(self.showSettings)

        settingsMenu.addSeparator()

        delConfirmation = settingsMenu.addAction(_('Delete items without confirmation:'))

        def switch_del_confirmation():
            settings = QtCore.QSettings()
            value = settings.value('tree/no_confirm_del', type=bool)
            settings.setValue('tree/no_confirm_del', not value)

        settings = QtCore.QSettings()
        value = settings.value('tree/no_confirm_del', type=bool)
        switch_del_confirmation.value = value

        delConfirmation.setCheckable(True)
        delConfirmation.setChecked(value)
        delConfirmation.triggered.connect(switch_del_confirmation)

        helpMenu = menu.addMenu(_("Help"))
        keysAction = helpMenu.addAction(_("Hot keys"))
        keysAction.triggered.connect(self.showHotkeys)

        if DEBUG:
            debugMenu = menu.addMenu(_("&Debug"))
            winSize = debugMenu.addAction(_("Change to fixed window size"))
            winSize.triggered.connect(self.setWinSize)

            printLog = debugMenu.addAction(_("Print log to console in plain text"))
            printLog.triggered.connect(self.printLog)

            showButtons = debugMenu.addAction(_("Show buttons"))
            showButtons.triggered.connect(self.cfgTree.buttons.show)

            selectCfg = debugMenu.addAction(_("show select config dialog"))
            selectCfg.triggered.connect(self.selectConfig)

            dumpCfg = debugMenu.addAction(_("Print config"))
            dumpCfg.triggered.connect(self.dumpConfig)

            testMonitoring = debugMenu.addAction(_("Send event to monitoring window"))
            testMonitoring.triggered.connect(self.testMonitoring)

            updateMonitoring = debugMenu.addAction(_("Update monitoring window"))
            updateMonitoring.triggered.connect(self.updateMonitoring)

            getLogs = debugMenu.addAction(_("Request logs"))
            getLogs.triggered.connect(self.getLogs)

            disableLocalhost = debugMenu.addAction(_("Disable localhost"))
            disableLocalhost.triggered.connect(self.disableLocalhost)

            printSVG = debugMenu.addAction(_("printSVG"))
            printSVG.triggered.connect(self.printSVG)

            printCRC = debugMenu.addAction(_("Print cfg's CRC"))
            printCRC.triggered.connect(self.printCRC)

            printStateDict = debugMenu.addAction("Print state_dict")
            printStateDict.triggered.connect(self.printStateDict)

        self.setMenuBar(menu)

    def editEnable(self, enable):
        self.copyAction.setEnabled(enable)
        self.pasteAction.setEnabled(enable)
        self.addVK.setEnabled(enable)
        self.addTest.setEnabled(enable)
        self.delVK.setEnabled(enable)
        self.delTest.setEnabled(enable)
        self.renameAction.setEnabled(enable)

    def updateRecentMenu(self):
        self.recentMenu.clear()
        try:
            with open(os.path.expanduser('~/.stp/ini')) as f:
                j = json.load(f)
                files = j['Path to open cfg']
        except:
            files = []

        if not files:
            self.recentMenu.setEnabled(False)
        else:
            self.recentMenu.setEnabled(True)
            for f in files:
                action = self.recentMenu.addAction(f)

                def func(x):
                    path = str(self.sender().text())
                    self.openFile(path)

                action.triggered.connect(func)

    def initUI(self):
        """Init User Interface"""
        self.resize(1024, 768)
        self.createPopupMenu = lambda: None
        centralWidget = QtWidgets.QWidget()
        upperWidget = QtWidgets.QWidget()
        hbox = QtWidgets.QHBoxLayout()
        vbox = QtWidgets.QVBoxLayout()
        upperWidget.setLayout(hbox)
        self.logView = LogViewer(DEBUG)
        vsplitter = QtWidgets.QSplitter()
        vsplitter.addWidget(upperWidget)
        vsplitter.setOrientation(Qt.Vertical)
        vsplitter.addWidget(self.logView)
        vsplitter.setStretchFactor(0, 4)
        vsplitter.setStretchFactor(1, 1)
        vbox.addWidget(vsplitter)

        centralWidget.setLayout(vbox)

        self.cfgTree = TreeWidget(self)

        self.cfgTabs = QtWidgets.QTabWidget()
        self.logPage = ResultsWidget(self)
        scrollLayout = QtWidgets.QVBoxLayout()
        self.scrollArea = QtWidgets.QScrollArea()
        self.scrollArea.setWidgetResizable(True)
        self.propsPage = PropertiesWidget(self)
        self.scrollArea.setWidget(self.propsPage)
        scrollLayout.addWidget(self.scrollArea)
        scrollLayout.addWidget(self.propsPage.buttons)
        self.propGrid = QtWidgets.QGridLayout()
        propsLayout = QtWidgets.QVBoxLayout()
        propsLayout.addLayout(self.propGrid)
        self.tw = QtWidgets.QTabWidget()

        self.tw.setTabShape(QtWidgets.QTabWidget.Triangular)
        self.tw.setStyleSheet("QTabWidget::pane { border: 1px solid; }")

        self.tw.hide()
        propsLayout.addWidget(self.tw)  # <---- here
        instTabs = [str(x) for x in range(1, 17)]
        self.tabCombo = QtWidgets.QComboBox()
        self.tabCombo.addItems(instTabs)
        self.tabCombo.currentIndexChanged.connect(self.setInstancesNumber)
        self.tw.setCornerWidget(self.tabCombo, Qt.TopLeftCorner)
        # propsLayout.addLayout(self.instanceLayout)
        self.propsPage.setLayout(propsLayout)
        widget = QtWidgets.QWidget()
        widget.setLayout(scrollLayout)
        self.cfgTabs.addTab(widget, _("Properties"))
        page = QtWidgets.QWidget()

        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addWidget(QtWidgets.QLabel(_('Auto-update results:')))
        self.log_checkbox = QtWidgets.QCheckBox()
        self.log_checkbox.setChecked(True)

        def f(state):
            if state:
                self.logPage.timer.start()
            else:
                self.logPage.timer.stop()

        self.log_checkbox.stateChanged.connect(f)

        button_layout.addWidget(self.log_checkbox)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.logPage)
        layout.addLayout(button_layout)
        page.setLayout(layout)
        self.cfgTabs.addTab(page, _("Results"))

        self.cfgTabs.currentChanged.connect(lambda: self.logPage.displayLogs(
            self.cfgTree.currentItem(), True))

        self.hsplitter = QtWidgets.QSplitter()
        self.hsplitter.addWidget(self.cfgTree)
        self.hsplitter.addWidget(self.cfgTabs)
        hbox.addWidget(self.hsplitter)
        self.setCentralWidget(centralWidget)

        # Add menu and actions
        self.initMenu()
        # Add toolbar
        self.toolbar = QtWidgets.QToolBar()
        self.toolbar.setMovable(False)
        self.toolbar.addAction(self.newFileAction)
        self.toolbar.addAction(self.openAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.connectTestsAction)
        self.toolbar.addAction(self.disconnectTestsAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.runTestsAction)
        self.toolbar.addAction(self.stopTestsAction)
        self.toolbar.addAction(self.pauseTestsAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.resetTestsAction)
        self.toolbar.addSeparator()
        self.statusLabel = QtWidgets.QLabel('')
        f = QFont("Arial", 12)

        self.statusLabel.setFont(f)
        self.toolbar.addWidget(self.statusLabel)
        self.addToolBar(self.toolbar)
        self.cfgTree.currentItemChanged.connect(self.showNode)
        self.cfgTree.currentItemChanged.connect(self.logPage.displayLogs)

        # Add status bar
        self.setStatusBar(statusbar(self))

        # Set focus to tree widget
        self.cfgTree.setFocus()

    def initLogger(self):
        handler = logHandler()
        handler.setLogWidget(self.logView)

        logger = logging.getLogger()
        logger.addHandler(handler)

    def initTimer(self):
        self._timer = QtCore.QTimer()
        self._timer.start(1000)
        self._timer.timeout.connect(self.statusBar().update_time)

    def __init__(self,cfg=None,host=None,debug=None, parent = None, skip_status=False, exclusive=False , autostart=False, session=False, english=False):
        super(GUIQT, self).__init__(parent)
        if self.__initialized:
            return

        if debug:
            global DEBUG
            DEBUG = True
            # self.setFixedSize(800, 600)
            self.debug("Running in debug mode")

        QtCore.QCoreApplication.setOrganizationName('mcst')
        QtCore.QCoreApplication.setApplicationName('stp/stp_gui')
        QtWidgets.QMainWindow.__init__(self, parent)
        QtWidgets.QApplication.setStyle("QtCurve")
        if os.path.exists(Common.data_dir + '/main.png'):
            self.setWindowIcon(QIcon(Common.data_dir + '/main.png'))
        else:
            self.setWindowIcon(
                QtWidgets.qApp.style().standardIcon(
                    QtWidgets.QStyle.SP_DesktopIcon))
        windowtitle.__init__(self)   #windowtitle.__init__(self,self)
        self.windowtitle_update()
        # TODO: Ui_Form is the class name generated by pyuic4
        # self.ui = Ui_Form()
        # self.ui.setupUi(self)
        self.stpCtrl = StpControl(self)
        self.stpCtrl.skip = skip_status
        # self.worker_thread = QtCore.QThread()
        # self.stpCtrl.moveToThread(self.worker_thread)
        # self.worker_thread.start()

        self.schemaparser = self.stpCtrl.get_schema_parser()

        # ID of client in exclusive mode
        self._excl_id = -1
        #print(self.stpCtrl.heartBeat)
        self.current_state = None
        self.old_state = None
        self.buttons_locked = False
        self.status_locked = False
        self.initUI()
        self.userModified = False

        self.elementsLock()
        self.initTimer()
        self.initLogger()

        self._hostname = 'localhost'
        self._port = '9876'

        self.stpCtrl.heartBeat.connect(self.update_gui)
        self.stpCtrl.changeState.connect(self.statusBar().update)
        self.stpCtrl.heartBeat.connect(self.statusBar().update_heartbeat)
        self.stpCtrl.heartBeat.connect(self.cfgTree.updateStatus)
        self.stpCtrl.changeState.connect(self.elementsLock)
        self.stpCtrl.heartBeat.connect(self.elementsLock)
        self.stpCtrl.diffConfig.connect(self.receiveChanges)
        self.stpCtrl.exclusive_signal.connect(self.updateExclusive)
        self.stpCtrl.logReceived.connect(self.logPage.showLog)
        self.stpCtrl.infoReceived.connect(self.logPage.showInfo)
        self.stpCtrl.editor_callbacks.append(self.disconnectButtonsDisable)
        self.update_gui()

        self.__initialized = True

        self.statusBar().update()

        self._currTab = None
        if cfg is not None:
            self.openFile(cfg)

        if host is not None:
            self._netConnect(host)
            if exclusive:
                self.stpCtrl.stp_setexcl(excl=True)
            if autostart:
                self.stpCtrl.stp_setautostart(flag=True)

        if not WEBKIT_INSTALLED:
            msg = "No QtWebKit installed. Monitoring window disabled."
            logging.warning(_(msg))

    # ------------------------------------
    # TREE PART / PROPS EDITOR PART
    # ------------------------------------

    # TODO: нужно как-то перегруппировать функции для наглядности
    # и добавить краткое описание функций
    def showNode(self, curr=None, prev=None):
        """Display properties for node selected in tree"""
        self.setUpdatesEnabled(False)
        if self.userModified:
            #button = QtWidgets.QMessageBox.question(
            #    self, _('Settings changed'),
            #    _('Leave page? Settings will be ignored'),
            #    QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Cancel)
            #if button == QtWidgets.QMessageBox.Cancel:
            #    self.cfgTree.currentItemChanged.disconnect(self.showNode)
            #    self.cfgTree.setCurrentItem(prev)
            #    self.cfgTree.currentItemChanged.connect(self.showNode)
            #    return
            #else:
            self.rejectChanges()
        if len(self.cfgTree.selectedItems()) == 0:
            if not self.cfgTree.currentItem():
                return
            path = str(
                self.cfgTree.currentItem().data(
                    0, Qt.UserRole)) #.toString()
            node = self.cfgTree.getNodeByPath(path)
            node.setSelected(True)
            self.showNode()
            self.setUpdatesEnabled(True)
            return
        elif len(self.cfgTree.selectedItems()) > 1:
            self.setUpdatesEnabled(True)
            return
        self.testCopy = None
        path = str(
            self.cfgTree.currentItem().data(
                0, Qt.UserRole))
        info = path[1:].split('/')
        d = {}
        fields = ['host', 'test', 'instance']
        for i, f in zip(info, fields):
            d[f] = i
        cfg = self.stpCtrl.get_config()

        self.level = (lambda: len(d) if d['host'] != '' else 0)()
        self.tw.hide()
        if self.level == 0:  # root node
            self.propsPage.fillTableCells(
                cfg.options.data, self.schemaparser.rootProperties, level=0)
            for w in self.logPage.widgets:
                w.setParent(None)
        elif self.level == 1:  # host node
            hosts = cfg.hosts
            host = next(
                (host for host in hosts if host.hostname == d['host']), None)
            if host:
                self.propsPage.fillTableCells(
                    host.options.data,
                    self.schemaparser.hostOptions,
                    level=1, hostname=host.hostname)
            for w in self.logPage.widgets:
                w.setParent(None)
        elif self.level == 2:  # test node
            for i in range(self.tw.count()):
                if self.tw.widget(i) is not None:
                    self.tw.widget(i).setParent(None)
            self.tw.clear()
            hosts = cfg.hosts
            host = next(
                (host for host in hosts if host.hostname == d['host']), None)
            tests = host.tests
            test = next(
                (test for test in tests if test.testname == d['test']), None)
            ctr = 0
            for inst in test.instances:
                options = inst.data
                options_param = options['param'].data
                l = QtWidgets.QGridLayout()
                w = QtWidgets.QWidget()
                w.setLayout(l)
                self.tw.addTab(w, 'instance ' + str(ctr))
                self.propsPage.fillTableCells(
                    options,
                    self.schemaparser.instanceOptions,
                    grid=l,
                    level=3,
                    instNum=ctr)
                self.propsPage.fillTableCells(
                    options_param, self.schemaparser.paramOptions, delta=1,
                    grid=l, level=4, instNum=ctr)
                ctr += 1
            self.propsPage.fillTableCells(
                test.data,
                self.schemaparser.testOptions,
                grid=self.propGrid,
                delta=2,
                level=2)
            self.tw.show()
            if self._currTab is not None:
                self.tw.setCurrentIndex(self._currTab)
            self.tabCombo.currentIndexChanged.disconnect(
                self.setInstancesNumber)
            self.tabCombo.setCurrentIndex(self.tw.count() - 1)
            self.tabCombo.currentIndexChanged.connect(self.setInstancesNumber)
        self.setUpdatesEnabled(True)

    def setInstancesNumber(self, index):
        self.propsPage.buttons.show()
        wants = index + 1
        have = self.tw.count()
        cfg = self.stpCtrl.get_config()
        path = str(self.cfgTree.currentItem().data(
            0, Qt.UserRole))  #.toString()
        hostname = path.split('/')[1]
        testname = path.split('/')[2]
        hosts = cfg.hosts
        host = next((host for host in hosts if host.hostname == hostname), None)
        tests = host.tests
        # test = next((test for test in tests if test.testname == testname), None)
        # testCopy = deepcopy(test)
        if self.testCopy is None:
            self.testCopy = self.restoreTestState()
        else:
            self.testCopy = self.restoreTestState(self.testCopy)
        if have < wants:
            for i in range(have, wants):
                previous_inst = self.testCopy.instances[have - 1]
                inst = instance()
                inst.data = deepcopy(previous_inst.data)
                options = inst.data
                options_param = options['param'].data
                l = QtWidgets.QGridLayout()
                w = QtWidgets.QWidget()
                w.setLayout(l)
                self.tw.addTab(w, 'instance ' + str(i))
                self.propsPage.fillTableCells(
                    options, self.schemaparser.instanceOptions, grid=l, level=3, instNum=i)
                self.propsPage.fillTableCells(
                    options_param, self.schemaparser.paramOptions, delta=1,
                    grid=l, level=4, instNum=i)
                self.testCopy.instances.append(inst)

                for x in previous_inst.data:
                    if x in ['param']:
                        for y in previous_inst[x].data:
                            self.propsPage.modifyConf(previous_inst[x].data[y], y, 4, i)
                            inst[x].data[y] = previous_inst[x].data[y]
                        continue
                    self.propsPage.modifyConf(previous_inst[x], x, 3, i)
                    inst.data[x] = previous_inst[x]

            self.propsPage.fillTableCells(
                self.testCopy.data, self.schemaparser.testOptions,
                grid=self.propGrid, delta=2, level=2)
        elif have > wants:
            # print 'have: ', have, ',wants: ', wants
            for i in range(have, wants, -1):
                self.testCopy.instances.pop()
                self.tw.removeTab(self.tw.count() - 1)
            # Empty record in modificationsQueue will force
            # sending MDF_CHN_TEST command
            self.propsPage.modifyConf(None, None, None, None)

        self.cfgTree.setDisabled(True)

    # ------------------------------------
    # SLOTS - process signals
    # ------------------------------------

    def update_gui(self, data=None):
        """ обработчик события об изменении состояния StpControl """

        if self.status_locked:
            return

        # TODO: ниже нужно превратить в статусную строку с отдельным полем
        # для каждого флага
        #
        text = _("Test status: ")
        if (data is None) or not self.stpCtrl.connected:
            msg = _("No connection")
            text += "<font color='black'>" + msg + "</font>"
        else:
            state = json.loads(str(data))['system']['state']
            if state & 1:
                msg = _('Running')
                text += "<font color='green'>" + msg + "</font>"
            elif state & 16:
                msg = _('Not Running')
                text += "<font color='red'>" + msg + "</font>"
            elif state & 2:
                msg = _('Pause')
                text += "<font color='red'>" + msg + "</font>"
        self.statusLabel.setText(text)
        # self.debug("GUI: Update GUI")
        self.windowtitle_update_run(msg)

    def closeEvent(self, event):
        """Ask user if he wants to stop kernel and monitors, then quit"""

        if self.propsPage.buttons.isVisible():
            reply = QtWidgets.QMessageBox.question(
                self, _('Config modified'),
                _('Save changes in config?'),
                QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)  # QtWidgets.QMessageBox.Cancel
            if reply == QtWidgets.QMessageBox.Cancel:
                event.ignore()
                return
            elif reply == QtWidgets.QMessageBox.Yes:
                self.applyChanges()
                self.saveFile()
            elif reply == QtWidgets.QMessageBox.No:
                self.rejectChanges()

        dialog = CloseDialog()
        dialog.exec_()
        ok = bool(dialog.result())
        if not ok:
            event.ignore()
            return

        item = dialog.get_item()

        settings = QtCore.QSettings()
        settings.setValue('last_session/exit_type', item)

        if item == 'Terminate GUI and kernel':
            self.stpCtrl.stp_disconnect(kernclose=True, monclose=False)
            event.accept()
        elif item == 'Terminate GUI only':
            if self.stpCtrl.connected:
                self.stpCtrl.stp_disconnect(kernclose=False, monclose=False)
            event.accept()
        elif item == 'Terminate all':
            self.stpCtrl.stp_disconnect(kernclose=True, monclose=True)
            event.accept()

        if self.monitoringWin is not None:
            self.monitoringWin.closeMonitoring()

        if self.hotkeys_win is not None:
            self.hotkeys_win.close()
        # sys.exit(0)

    # ------------------------------------
    # MAIN MENU ACTIONS
    #   File | Edit | View | Tests
    # ------------------------------------

    # File: New | Open | Save | Save As | Close | Quit

    def newFile(self):
        """Create new config file"""
        self.stpCtrl.config_create()
        # TODO We shouldn't check if we are connected here, but config_create
        # works different from other actions
        if not self.stpCtrl.connected:
            self.cfgTree.displayCfg(newCfg=True)

    def openFile(self, filename=None):
        """Load config file"""
        if self.userModified:
            button = QtWidgets.QMessageBox.question(
                self, _('Config modified'),
                _('Open file? changes will be lost'),
                QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Cancel)
            if button == QtWidgets.QMessageBox.Cancel:
                return
            self.rejectChanges()
        if filename is None:
            bufferfilename = QtWidgets.QFileDialog.getOpenFileName(
                self, _('Open file'),
                '', _('STP Configuration Files (*.cfg);;All Files (*)'), None,
                QtWidgets.QFileDialog.DontUseNativeDialog)
            filename = bufferfilename[0]
        else:
            if not os.path.isfile(filename):
                logging.warning(_("Incorrect path to configuration file: ") + filename)
                return
        if not filename:
            return
        try:
            self.stpCtrl.config_read(filename)
        except ValueError:
            logging.error(_('Configuration file is not a valid JSON'))
            errWin = QtWidgets.QMessageBox()
            errWin.critical(
                None, _('Error'),
                _('Configuration file is not a valid JSON'))
            errWin.show()
            return False
        logging.info(_("Configuration loaded from: ") + str(filename))
        self.cfgTree.displayCfg(newCfg=True)
        self.addRecentFile(filename)

    def addRecentFile(self, filename):
        # Save path to ini file
        if filename[0] != '/':
            filename = os.path.abspath(str(filename))
        try:
            with open(os.path.expanduser('~/.stp/ini')) as f:
                j = json.load(f)

            try:
                files = j['Path to open cfg']
            except KeyError:
                # If we can open file, but there are no 'Path to open cfg' entries
                j['Path to open cfg'] = []
                files = j['Path to open cfg']

            if filename not in files:
                files.append(str(filename))
            with open(os.path.expanduser('~/.stp/ini'), 'w') as f:
                json.dump(j, f, indent=4)
            self.updateRecentMenu()
        except Exception as e:
            self.warning('Error when writing settings to ~/.stp/ini' + str(e))

    def saveFile(self):
        """Save current config"""
        if not self.stpCtrl.config_present():
            logging.info(_("No configuration to save"))
            return
        filename = self.stpCtrl.config_getfilename()
        if not filename or filename == '/':
            self.saveFileAs()
            return
        with open(filename, 'r') as f:
            if f.readline().startswith('STP'):
                button = SimpleConfigDialog().exec_()
                if button == QtWidgets.QMessageBox.Save:
                    self.saveFileAs()
                    return
                elif button == QtWidgets.QMessageBox.Cancel:
                    return

        if not self.stpCtrl.config_write():
            logging.error(_('Error while saving file'))
        else:
            self.addRecentFile(filename)

    def saveFileAs(self):
        """Save current config as ..."""
        if not self.stpCtrl.config_present():
            logging.info(_("No configuration to save"))
            return
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, _('Save file as...'),
            '', _('STP Configuration Files (*.cfg);;All Files (*)'), None,
            QtWidgets.QFileDialog.DontUseNativeDialog)[0]
        if not filename:
            return
        if not self.stpCtrl.config_write(filename):
            #            logging.error(_('Error while saving file'))
            return
        else:
            self.addRecentFile(filename)

    def closeFile(self):
        """Close loaded config"""
        self.stpCtrl.close_cfg()

    # Tests: Connect | Disconnect | Run | Stop | Pause | Reset | Exclusive

    def _netConnect(self, url):
        url_split = str(url).split(':')
        if len(url_split) == 2:
            self._hostname = url_split[0]
            self._port = url_split[1]
            self.stpCtrl.stp_connect(url_split[0], url_split[1])
        elif len(url_split) == 1:
            self._hostname = url_split[0]
            self.stpCtrl.stp_connect(url_split[0], "9876")
            try:
                hosts = []
                with open(os.path.expanduser('~/.stp/ini')) as f:
                    j = json.load(f)
            except IOError:
                default = '{"Address for host connect": [],' \
                          ' "Path to save cfg":[], "Path to open cfg": []}'
                j = json.loads(default)

            try:
                try:
                    hosts = j['Address for host connect']
                except KeyError:
                    # If we can open file, but there are no 'Address for host connect' entries
                    j['Address for host connect'] = []
                    hosts = j['Address for host connect']

                if self._hostname in hosts:
                    hosts.remove(self._hostname)

                hosts.append(self._hostname)

                dir_path = os.path.expanduser('~/.stp')
                if not os.path.exists(dir_path):
                    os.makedirs(dir_path)
                with open(os.path.expanduser('~/.stp/ini'), 'w+') as f:
                    json.dump(j, f, indent=4)
            except ValueError:
                # If we can open file, but can't parse json
                return
            except IOError:
                # If we can't open file
                return

    def netConnect(self):
        """Try connect to STP kernel"""
        # Request hostname:port from user
        hosts = []
        try:
            with open(os.path.expanduser('~/.stp/ini')) as f:
                j = json.load(f)
            hosts = list(reversed(j['Address for host connect']))
        except KeyError:
            # if we can open file, but no entries exists
            hosts.append(self._hostname)
        except ValueError:
            # If we can open file, but can't parse json
            hosts.append(self._hostname)
        except IOError:
            # If we can't open file
            hosts.append(self._hostname)

        if self._port == "9876":
            dialog = ConnectDialog(hosts)
        else:
            suggestions = ['%s:%s' % (self._hostname, self._port)] + hosts
            dialog = ConnectDialog(suggestions)
        dialog.exec_()

        ok = dialog.result()
        url = dialog.textValue()

        if not ok or not url:
            return

        self.log_checkbox.setChecked(True)
        self._netConnect(url)

    def chooseStyle(self):
        styles = []
        for key in QtWidgets.QStyleFactory.keys():
            st = QtWidgets.QStyleFactory.create(key)
            # print(key, str(key), st.metaObject().className())
            styles.append(str(key))

        dialog = StyleDialog(styles)
        dialog.exec_()

        ok = dialog.result()
        style = dialog.textValue()

        if not ok or not style:
            return

        self.info(_('Chosen style') + ': ' + style)
        QtWidgets.QApplication.setStyle(style)

    def startKernel(self):
        subprocess.Popen([Common.bin_dir + '/stp_kernel', '--alone'])
        time.sleep(1)
        self._netConnect('localhost')

    def netDisconnect(self):
        """Disonnect from STP kernel"""
        self.stpCtrl.stp_disconnect(kernclose=False, monclose=False)

    def netReconnect(self):
        """Reconnect all monitors with kernel"""
        self.stpCtrl.stp_reconnectmon()

    def runTests(self):
        """Run selected tests"""
        self.buttonsLock()
        # print 'RUN'
        self.stpCtrl.test_start()

    def stopTests(self):
        """Stop selected tests"""
        # print 'STOP'
        self.buttonsLock()
        self.stpCtrl.test_stop()

    def pauseTests(self):
        """Pause tests"""
        # print 'PAUSE'
        self.buttonsLock()
        self.stpCtrl.test_pause()

    def resetTests(self):
        """ Reset statistics """
        if self.stpCtrl.connected:
            self.stpCtrl.test_reset()
        else:
            self.cfgTree.resetTestsData()

    def isReadOnlyMode(self):
        if self.stpCtrl.exclusive:
            return False
        if self._excl_id != self.stpCtrl.stp_id and self._excl_id != -1 and self.stpCtrl.connected:
            return True
        return False

    def exclusiveMode(self):
        """ Request exclusive mode from kernel """
        if self.stpCtrl.exclusive:
            self.stpCtrl.stp_setexcl(excl=False)
        else:
            self.stpCtrl.stp_setexcl(excl=True)

    def getTestLogs(self, host, test, instance):
        self.stpCtrl.get_log(host, test, instance)

    def getHostInfo(self, host):
        self.stpCtrl.get_info(host)

    # Edit: Undo | Redo | Copy | Paste | Add* | Del* | Rename

    def undo(self):
        pass

    def redo(self):
        pass

    def copy(self):
        pass

    def paste(self):
        pass

    # View: Select | Deselect | Display | Max Tree | Max Options

    def selectAll(self):
        self.stpCtrl.mdf_testall_on()

    def deselectAll(self):
        self.stpCtrl.mdf_testall_off()

    def toggleNode(self):
        """ Display """
        pass

    def maxTree(self):
        self.hsplitter.setSizes([1, 0])

    def maxCfg(self):
        self.hsplitter.setSizes([0, 1])

    # ------------------------------------
    # ACTIONS
    # Functions for signals from Apply and Cancel buttons
    # ------------------------------------

    def restoreTestState(self, test=None):
        if test is None:
            cfg = self.stpCtrl.get_config()
            path = str(self.cfgTree.currentItem().data(
                0, Qt.UserRole)) #.toString()
            hostname = path.split('/')[1]
            testname = path.split('/')[2]
            hosts = cfg.hosts
            host = next(
                (host for host in hosts if host.hostname == hostname), None)
            tests = host.tests
            test = next(
                (test for test in tests if test.testname == testname), None)
        cfgPartCopy = deepcopy(test)

        for entry in self.propsPage.mods.entries.values():
            if entry.level == 2:  # we are in test node
                if entry.index is None:  # If entry is not combobox entry
                    setattr(cfgPartCopy, entry.key, entry.value)
                else:
                    setattr(cfgPartCopy, entry.key, entry.index[entry.value])
            elif entry.level == 3:  # we are in instance node
                if entry.instance_num >= len(test.instances):
                    continue
                if entry.index is None:
                    setattr(
                        cfgPartCopy.instances[entry.instance_num],
                        entry.key,
                        entry.value)
                else:
                    setattr(
                        cfgPartCopy.instances[
                            entry.instance_num], entry.key, entry.index[
                            entry.value])
            elif entry.level == 4:  # we are in utest options node
                if entry.instance_num >= len(test.instances):
                    continue
                if entry.index is None:
                    setattr(
                        cfgPartCopy.instances[
                            entry.instance_num].data['param'],
                        entry.key,
                        entry.value)
                else:
                    setattr(
                        cfgPartCopy.instances[
                            entry.instance_num].data['param'], entry.key, entry.index[
                            entry.value])
        return cfgPartCopy

    def applyChanges(self):
        if self._excl_id == -1 or self.stpCtrl.stp_id == -1:
            pass
        elif self._excl_id != self.stpCtrl.stp_id:
            self.propsPage.buttons.hide()
            self.userModified = False
            # self.stpCtrl.diffConfig.connect(self.receiveChanges)
            self.cfgTree.displayCfg()
            self.propsPage.mods.clear()
            # self.cfgTree.currentItemChanged.connect(self.logPage.displayLogs)
            self.cfgTree.setDisabled(False)
            self.toolbar.setDisabled(False)
            self.menuBar().setDisabled(False)
            return

        # self.cfgTree.currentItemChanged.disconnect(self.logPage.displayLogs)
        """ [Apply] Button """
        path = str(self.cfgTree.currentItem().data(
            0, Qt.UserRole))
        cfg = self.stpCtrl.get_config()
        self._currTab = self.tw.currentIndex()

        # if user only changed number of instances:
        if len(self.propsPage.mods.entries) == 0 and self.level == 2:
            hostname = path.split('/')[1]
            testname = path.split('/')[2]
            hosts = cfg.hosts
            host = next(
                (host for host in hosts if host.hostname == hostname), None)
            tests = host.tests
            test = next(
                (test for test in tests if test.testname == testname), None)
            self._e = deepcopy(test)
            if self.testCopy is not None:
                while len(self._e.instances) < len(self.testCopy.instances):
                    self._e.instances.append(instance())
                while len(self._e.instances) > len(self.testCopy.instances):
                    self._e.instances.pop()
            # self.stpCtrl.mdf_chntest(hostname, testname, t.dump(compact=True))
        flag = True
        for entry in self.propsPage.mods.entries.values():
            if self.level == 0:  # Root
                cfgPart = cfg.options
                if flag:
                    flag = False
                    self._e = deepcopy(cfgPart)
                if entry.index is None:
                    setattr(self._e, entry.key, entry.value)
                else:
                    setattr(self._e, entry.key, entry.index[entry.value])
                # self.stpCtrl.mdf_chnpref(self._e.dump(compact=True))

            elif self.level == 1:  # Host
                hostname = path.split('/')[1]
                hosts = cfg.hosts
                host = next(
                    (host for host in hosts if host.hostname == hostname), None)
                if flag:
                    flag = False
                    self._e = deepcopy(host)
                cfgPart = self._e.options
                if entry.index is None:
                    setattr(cfgPart, entry.key, entry.value)
                else:
                    setattr(cfgPart, entry.key, entry.index[entry.value])
                # self.stpCtrl.mdf_chnvk(hostname, self._e.dump(compact=True))

            elif self.level == 2:  # Test
                hostname = path.split('/')[1]
                testname = path.split('/')[2]
                hosts = cfg.hosts
                host = next(
                    (host for host in hosts if host.hostname == hostname), None)
                tests = host.tests
                test = next(
                    (test for test in tests if test.testname == testname), None)
                # We need to make sure there are equal number of instances in
                # self.testCopy and in test TODO: do it outside entries
                # iteration
                if self.testCopy is not None:
                    while len(test.instances) < len(self.testCopy.instances):
                        test.instances.append(instance())
                    while len(test.instances) > len(self.testCopy.instances):
                        test.instances.pop()
                # if entry.instance_num is None: # we are in test node
                if flag:
                    flag = False
                    self._e = deepcopy(test)
                if entry.level == 2:  # we are in test node
                    if entry.index is None:
                        setattr(self._e, entry.key, entry.value)
                    else:
                        setattr(
                            self._e,
                            entry.key,
                            entry.index[
                                entry.value])
                elif entry.level == 3:  # we are in instance node
                    # Skip changes in removed instances
                    if entry.instance_num < len(self._e.instances):
                        if entry.index is None:
                            setattr(
                                self._e.instances[entry.instance_num],
                                entry.key,
                                entry.value)
                        else:
                            setattr(
                                self._e.instances[
                                    entry.instance_num], entry.key, entry.index[
                                    entry.value])
                elif entry.level == 4:  # we are in utest options node
                    if entry.instance_num < len(self._e.instances):
                        if entry.index is None:
                            setattr(
                                self._e.instances[entry.instance_num].data['param'],
                                entry.key,
                                entry.value)
                        else:
                            setattr(
                                self._e.instances[entry.instance_num].data['param'],
                                entry.key,
                                entry.index[entry.value])

                # self.stpCtrl.mdf_chntest(
                # hostname, testname, self._e.dump(
                # compact=True))

            # print "DUMP: ", self._e.dump(compact=True)
        self.cfgTree.setDisabled(False)
        self.toolbar.setDisabled(False)
        self.menuBar().setDisabled(False)

        if self.propsPage.mods.entries:
            if self.level == 0:
                if Config.closed:
                    self.stpCtrl.config_create()

                self.stpCtrl.mdf_chnpref(self._e.dump(compact=True))
            elif self.level == 1:
                self.stpCtrl.mdf_chnvk(hostname, self._e.dump(compact=True))
            elif self.level == 2:
                self.stpCtrl.mdf_chntest(hostname, testname, self._e.dump(compact=True))

        if self.propsPage.name_edit_text:
            name = self.propsPage.name_edit_text
            s = str(self.cfgTree.currentItem().data(0, Qt.UserRole))  #.toString()
            hostname = s[1:]
            # print hostname
            if hostname != name:
                hosts = self.stpCtrl.get_config().data['hosts']
                if name not in [h.hostname for h in hosts]:
                    h = next((x for x in hosts if x.hostname == hostname), None)
                    new_h = deepcopy(h)
                    new_h.hostname = str(name)
                    try:
                        del self.cfgTree.ignore_dict['/' + str(hostname)]
                    except KeyError:
                        pass
                    s = str(self.cfgTree.currentItem().data(0, Qt.UserRole)) #.toString()
                    self.cfgTree.treeState.renamed(s, '/' + str(name))
                    self.stpCtrl.mdf_chnvk(hostname, new_h.dump())
                else:
                    logging.error("GUI: " + _("Node {} already exists").format(name))
            self.propsPage.name_edit_text = ''
        self.propsPage.buttons.hide()
        self.userModified = False
        # self.stpCtrl.diffConfig.connect(self.receiveChanges)
        self.cfgTree.displayCfg()
        self.propsPage.mods.clear()
        # self.cfgTree.currentItemChanged.connect(self.logPage.displayLogs)
        # self.stpCtrl.heartBeat.connect(self.cfgTree.updateStatus)#wtf

    def rejectChanges(self):
        """ [Cancel] Button """
        # self.cfgTree.displayCfg()
        self._currTab = self.tw.currentIndex()
        self.propsPage.mods.clear()
        self.userModified = False
        self.showNode()
        self.propsPage.buttons.hide()
        self.cfgTree.setDisabled(False)
        self.toolbar.setDisabled(False)
        self.menuBar().setDisabled(False)
        # self.stpCtrl.heartBeat.connect(self.cfgTree.updateStatus)#wtf

    def showTerminateDialog(self):
        dialog = TerminationDialog()
        dialog.exec_()
        ok = bool(dialog.result())
        if not ok:
            return
        item = dialog.get_item()

        if item == 'Terminate all':
            self.stpCtrl.stp_disconnect(kernclose=True, monclose=True)
        elif item == 'Terminate kernel only':
            self.stpCtrl.stp_disconnect(kernclose=True, monclose=False)

    def showSettings(self):
        self.settingsWin = SettingsWindow()
        self.settingsWin.show()

    def showMonitoring(self):
        try:
            os.remove('data/user-scheme.svg')
        except OSError:
            pass

        cfg = self.stpCtrl.get_config()

        # Use SVG from config if possible
        if cfg and cfg.svg:
            with open(os.path.join(Common.data_dir, 'user-scheme.svg'), 'w') as f:
                f.write(cfg.svg.replace('\\n', '\n').replace('\\"', '"'))

        if self.monitoringWin is None:
            self.monitoringWin = MonitoringWindow(self.cfgTree)

        if not self.monitoringWin:
            return None
        if DEBUG:
            self.monitoringWin.debug = True
        self.monitoringWin.show()

    def receiveChanges(self, cmd, message):
        self.debug('Cmd: ' + cmd)
        new_current_item = None
        if cmd == NetCmd.MDF_ADD_VK:
            json_dump = json.loads(str(message))
            host_dump = json_dump['vk'][0]

            h = host(host_dump)
            h.hostname = json_dump['vkname']
            cfg = self.stpCtrl.get_config()
            cfg.hosts.append(h)
            new_current_item = '/' + h.hostname
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_CHN_VK:
            json_dump = json.loads(str(message))
            host_dump = json_dump['vk'][0]
            old_name = json_dump['vkname']

            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts

            cfg = self.stpCtrl.get_config()
            for (i, vk) in enumerate(cfg.hosts):
                if vk.hostname == json_dump['vkname']:
                    cfg.hosts.pop(i)
                    break

            h = host(host_dump)
            cfg.hosts.append(h)
            new_current_item = '/' + str(h.hostname)
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_DEL_VK:
            json_dump = json.loads(str(message))
            name = json_dump['vkname']
            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            h = next((x for x in hosts if x.hostname == name), None)
            hosts.remove(h)

            node = self.cfgTree.getNodeByPath('/' + name)
            parent = node.parent()
            index = parent.indexOfChild(node)
            if index:
                node = parent.child(index - 1)
                self.cfgTree.setCurrentItem(node)
            else:
                node = parent.child(index + 1)
                if node:
                    self.cfgTree.setCurrentItem(node)
                else:
                    self.cfgTree.setCurrentItem(parent)
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_ADD_TEST:
            json_dump = json.loads(str(message))
            test_dump = json_dump['test'][0]
            t = test(test_dump)
            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            hname = json_dump['vkname']
            h = next((x for x in hosts if x.hostname == hname), None)
            tests = h.tests
            tests.append(t)
            new_current_item = '/' + h.hostname + '/' + t.testname
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_CHN_TEST:
            json_dump = json.loads(str(message))
            test_dump = json_dump['test'][0]
            t = test(test_dump)
            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            hname = json_dump['vkname']
            h = next((x for x in hosts if x.hostname == hname), None)
            tests = h.tests
            for (i, tst) in enumerate(tests):
                if tst.testname == json_dump['testname']:
                    tests.pop(i)
                    break

            tests.append(t)
            new_current_item = '/' + h.hostname + '/' + t.testname
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_DEL_TEST:
            json_dump = json.loads(str(message))
            hostname = json_dump['vkname']
            testname = json_dump['testname']
            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            h = next((x for x in hosts if x.hostname == hostname), None)
            tests = h.tests
            t = next((x for x in tests if x.testname == testname), None)
            tests.remove(t)

            node = self.cfgTree.getNodeByPath('/' + hostname + '/' + testname)
            parent = node.parent()
            index = parent.indexOfChild(node)
            if index:
                node = parent.child(index - 1)
                self.cfgTree.setCurrentItem(node)
            else:
                node = parent.child(index + 1)
                if node:
                    self.cfgTree.setCurrentItem(node)
                else:
                    self.cfgTree.setCurrentItem(parent)
            self.cfgTree.displayCfg()


        elif cmd == NetCmd.MDF_CHN_PREF:
            json_dump = json.loads(str(message))
            opt_dump = json_dump['options']
            opt = options(opt_dump)
            cfg = self.stpCtrl.get_config()
            cfg.options = opt
            cfg.data['options'] = opt
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.SET_CFG:
            cfg = Config()
            cfg.create(str(message))
            self.userModified = False
            self.cfgTree.displayCfg(newCfg=True)
            return

        elif cmd == NetCmd.MDF_VKALL_INVERT:
            cfg = self.stpCtrl.get_config()
            for (i, vk) in enumerate(cfg.hosts):
                if vk.options.enable:
                    vk.options.enable = 0
                else:
                    vk.options.enable = 1
            logging.info(_("All hosts was inverted."))
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_VKALL_ON:
            cfg = self.stpCtrl.get_config()
            for (i, vk) in enumerate(cfg.hosts):
                vk.options.enable = 1
            logging.info(_("All hosts was enabled."))
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_VKALL_OFF:
            cfg = self.stpCtrl.get_config()
            for (i, vk) in enumerate(cfg.hosts):
                vk.options.enable = 0
            logging.info(_("All hosts was disabled."))

        elif cmd == NetCmd.MDF_VK_ON:
            # TODO: make  tests of vk disabled as well
            json_dump = json.loads(str(message))

            cfg = self.stpCtrl.get_config()
            for (i, vk) in enumerate(cfg.hosts):
                if vk.hostname == json_dump['vkname']:
                    vk.options.enable = 1
                    logging.info(_("Host {} was enabled.").format(vk.hostname))
                    #self.cfgTree.displayCfg()
                    break

        elif cmd == NetCmd.MDF_VK_OFF:
            json_dump = json.loads(str(message))

            cfg = self.stpCtrl.get_config()
            for (i, vk) in enumerate(cfg.hosts):
                if vk.hostname == json_dump['vkname']:
                    vk.options.enable = 0
                    logging.info(_("Host {} was disabled.").format(vk.hostname))
                    self.cfgTree.displayCfg()
                    break

        elif cmd == NetCmd.MDF_TESTALL_INVERT:
            json_dump = json.loads(str(message))
            hostname = json_dump['vkname']

            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            h = next((x for x in hosts if x.hostname == hostname), None)
            tests = h.tests
            for (i, tst) in enumerate(tests):
                if tst.enable:
                    tst.enable = 0
                else:
                    tst.enable = 1
            logging.info(_("All tests on host {} was inverted.").format(hostname))
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_TESTALL_ON:
            json_dump = json.loads(str(message))
            hostname = json_dump['vkname']

            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            h = next((x for x in hosts if x.hostname == hostname), None)
            tests = h.tests
            for (i, tst) in enumerate(tests):
                tst.enable = 1
            logging.info(_("All tests on host {} was enabled.").format(hostname))
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_TESTALL_OFF:
            json_dump = json.loads(str(message))
            hostname = json_dump['vkname']

            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            h = next((x for x in hosts if x.hostname == hostname), None)
            tests = h.tests
            for (i, tst) in enumerate(tests):
                tst.enable = 0
            logging.info(_("All tests on host {} was disabled.").format(hostname))
            self.cfgTree.displayCfg()

        elif cmd == NetCmd.MDF_TEST_ON:
            json_dump = json.loads(str(message))
            hostname = json_dump['vkname']
            testname = json_dump['testname']

            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            h = next((x for x in hosts if x.hostname == hostname), None)
            tests = h.tests
            for (i, tst) in enumerate(tests):
                if tst.testname == testname:
                    tst.enable = 1
                    logging.info(_("Test {} on host {} was enabled.").format(testname, hostname))
                    break

        elif cmd == NetCmd.MDF_TEST_OFF:
            json_dump = json.loads(str(message))
            hostname = json_dump['vkname']
            testname = json_dump['testname']

            cfg = self.stpCtrl.get_config()
            hosts = cfg.hosts
            h = next((x for x in hosts if x.hostname == hostname), None)
            tests = h.tests
            for (i, tst) in enumerate(tests):
                if tst.testname == testname:
                    tst.enable = 0
                    logging.info(_("Test {} on host {} was disabled.").format(testname, hostname))
                    #self.cfgTree.displayCfg()
                    break

        elif cmd == NetCmd.CMD_RUN:
            # print 'RUN ACK'
            if message == '':
                self.buttonsUnlock()
            self.cfgTree.displayCfg()
        elif cmd == NetCmd.CMD_PAUSE:
            # print 'PAUSE ACK'
            if message == '':
                self.buttonsUnlock()
            self.cfgTree.displayCfg()
        elif cmd == NetCmd.CMD_STOP:
            # print 'STOP ACK'
            if message == '':
                self.buttonsUnlock()
            self.cfgTree.displayCfg()
        elif cmd == NetCmd.MES_MESSAGE:
            logging.info(_("Message from kernel: ") + _(str(message)))
            return
        self.userModified = False
        #self.cfgTree.displayCfg()
        if new_current_item:
            item = self.cfgTree.getNodeByPath(new_current_item)
            self.cfgTree.setCurrentItem(item)

    # ------------------------------------
    # DEBUG
    # Functions from debug menu
    # ------------------------------------

    def disconnectButtonsDisable(self):
        self.connectTestsAction.setEnabled(True)
        self.connectTestsAction.setVisible(True)
        self.disconnectTestsAction.setEnabled(False)
        self.disconnectTestsAction.setVisible(False)

        self.startKernelAction.setEnabled(True)
        self.startKernelAction.setVisible(True)

        self.exclusiveModeAction.setEnabled(False)
        self.terminateAction.setEnabled(False)
        self.reconnectAction.setEnabled(False)
        self.resetTestsAction.setEnabled(False)

        self.runTestsAction.setEnabled(False)
        self.stopTestsAction.setEnabled(False)
        self.resetTestsAction.setEnabled(False)
        self.pauseTestsAction.setEnabled(False)

    # it is half-durty hack, but work
    # todo: rewrite signals work
    def elementsLock(self):
        # text = 'elementsLock'
        if self.buttons_locked:
            # print text
            return
        self.propsPage.setEnabled(True)
        self.editEnable(True)
        if self.stpCtrl.connected:
            # text += ' connected'
            self.connectTestsAction.setEnabled(False)
            self.connectTestsAction.setVisible(False)
            self.disconnectTestsAction.setEnabled(True)
            self.disconnectTestsAction.setVisible(True)
            self.exclusiveModeAction.setEnabled(True)
            self.terminateAction.setEnabled(True)
            self.reconnectAction.setEnabled(True)
            self.resetTestsAction.setEnabled(True)
            self.startKernelAction.setEnabled(False)
            self.startKernelAction.setVisible(False)
        else:
            # text += ' disconnected'
            self.connectTestsAction.setEnabled(True)
            self.connectTestsAction.setVisible(True)
            self.disconnectTestsAction.setEnabled(False)
            self.disconnectTestsAction.setVisible(False)
            self.exclusiveModeAction.setEnabled(False)
            self.terminateAction.setEnabled(False)
            self.reconnectAction.setEnabled(False)
            self.resetTestsAction.setEnabled(True)
            self.startKernelAction.setEnabled(True)
            self.startKernelAction.setVisible(True)

        if self.stpCtrl.running:
            # text += ' running'
            self.runTestsAction.setEnabled(False)
            self.stopTestsAction.setEnabled(True)
            self.resetTestsAction.setEnabled(False)
            self.pauseTestsAction.setEnabled(True)
        elif self.stpCtrl.connected:
            # text += ' elserun_conn'
            self.runTestsAction.setEnabled(True)
            self.stopTestsAction.setEnabled(self.stpCtrl.paused)
            self.resetTestsAction.setEnabled(True)
            self.pauseTestsAction.setEnabled(False)
        else:
            # text += ' elserunning'
            self.runTestsAction.setEnabled(False)
            self.stopTestsAction.setEnabled(False)
            self.resetTestsAction.setEnabled(True)
            self.pauseTestsAction.setEnabled(False)

        if self.isReadOnlyMode():
            # text += ' RO'
            self.runTestsAction.setEnabled(False)
            self.stopTestsAction.setEnabled(False)
            self.resetTestsAction.setEnabled(False)
            self.pauseTestsAction.setEnabled(False)
            self.reconnectAction.setEnabled(False)
            self.exclusiveModeAction.setEnabled(False)
            self.terminateAction.setEnabled(False)
            self.terminateAction.setEnabled(False)
            self.propsPage.setEnabled(False)
            self.editEnable(False)
        # print text

    def buttonsLock(self):
        # print 'buttons locked'
        self.old_state = self.current_state
        self.buttons_locked = True
        self.runTestsAction.setEnabled(False)
        self.stopTestsAction.setEnabled(False)
        self.resetTestsAction.setEnabled(False)
        self.pauseTestsAction.setEnabled(False)

    def buttonsUnlock(self):
        # print 'buttons unlocked'
        self.buttons_locked = False
        self.status_locked = False

    def updateExclusive(self, excl):
        # not needed !
        self.statusBar().update()

    def selectConfig(self):
        dialog = SelectConfigDialog()
        if dialog.exec_() == QtWidgets.QMessageBox.Yes:
            self.stpCtrl.selectConfig = True
            # Replace config
            return True
        else:
            self.stpCtrl.selectConfig = False
            # Keep local config
            return False

    def showHotkeys(self):
        text = ''
        loc = locale.getlocale()
        if loc[0] == 'ru_RU' and loc[1] == 'KOI8-R':
            path = os.path.join(Common.doc_dir, 'gui_keys_ru.txt')
        else:
            path = os.path.join(Common.doc_dir, 'gui_keys.txt')

        try:
            with open(path, 'r') as f:
                text = f.read()
#               codec = QtCore.QTextCodec.codecForName("KOI8-R")
#               text = codec.toUnicode(text)
        except:
            text = _("Can't open help file")

        self.hotkeys_win = QtWidgets.QTextEdit()
        self.hotkeys_win.setPlainText(text)
        self.hotkeys_win.setReadOnly(True)

        document = self.hotkeys_win.document()
        font = document.defaultFont()
        font.setFamily('Courier New')
        document.setDefaultFont(font)

        font = self.hotkeys_win.document().defaultFont()
        fontMetrics = QFontMetrics(font)
        textSize = fontMetrics.size(0, text)

        textWidth = textSize.width() + 30
        textHeight = textSize.height() + 30

        self.hotkeys_win.setMinimumSize(textWidth, textHeight)
        self.hotkeys_win.resize(textWidth, textHeight)

        self.hotkeys_win.adjustSize()
        self.hotkeys_win.show()

    def setWinSize(self):
        self.setFixedSize(800, 600)

    def printLog(self):
        print(self.logView.toPlainText())

    def dumpConfig(self):
        cfg = Config()
        print(cfg.dump(compact=True, remdef=True))

    def testMonitoring(self):
        self.monitoringWin.changeColour()

    def updateMonitoring(self):
        s = "Ethernet: Err\\nCPU: OK\\nMEM: OK"
        self.monitoringWin.updateResults(s)

    def getLogs(self):
        self.stpCtrl.get_log('vk', 'test', 0)

    def disableLocalhost(self):
        # self.stpCtrl.mdf_vkon(on=False, vk='localhost')
        self.stpCtrl.mdf_teston(on=False, vk='localhost', test='test')

    def printSVG(self):
        cfg = Config()
        svg = cfg.svg
        print(svg)
        with open(os.path.join(Common.data_dir, 'user-scheme.svg'), 'w') as f:
            f.write(svg.replace('\\n', '\n').replace('\\"', '"'))

    def printCRC(self):
        cfg = Config()
        print(hex(cfg.crc()))

    def printStateDict(self):
        print(self.cfgTree.state_dict)

# ------------------------------------------------------------------------------
if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    gui = GUIQT()
    gui.show()
    sys.exit(app.exec_())
