#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Messages creater and parser

Like HTTP protocol using TCP/IP connection:
 - start line:
   STP <message_type> <message_answer>(empty-for-request)<CR><LF>
 - header (key: value):
   Size: 0<CR><LF>
   CRC: 0<CR><LF>
   Multiple: 0<CR><LF>
   Multi-Number: [1-]<CR><LF>
 - <CR><LF>
 - databody in JSON format.

message_type:
 - Information:
   MES_ACKNOWLEDGE MES_STATUS MES_SERVER_DONE MES_UI_DONE
 - Request:
   GET_CFG GET_STATUS GET_TST_STATUS GET_INFO_ABOUT_VK GET_TEST_LOG
   GET_DEV_LIST GET_ONLINE_IP_LIST GET_INFO_ABOUT_OSLTC
 - Setup commands:
   SET_CFG SET_EXCL SET_UN_EXCL SET_LOAD_SESSION
 - Test commands:
   CMD_RUN CMD_PAUSE CMD_CONTINUE CMD_STOP CMD_RESET CMD_RESTART
 - Control commands:
   WRK_TERMMONITOR WRK_RECONNECT WRK_DISCONNECT WRK_TERMINATE WRK_TERMALL
 - Change config commands:
   MDF_ADD_VK MDF_CHN_VK MDF_DEL_VK MDF_ADD_TEST MDF_CHN_TEST MDF_DEL_TEST
   MDF_VK_ON MDF_VK_OFF MDF_VKALL_INVERT MDF_VKALL_ON MDF_VKALL_OFF MDF_TEST_ON MDF_TEST_OFF
   MDF_TESTALL_INVERT MDF_TESTALL_ON MDF_TESTALL_OFF MDF_CHN_PREF

message_answer:
   RPL_OK RPL_BUSY RPL_NONE RPL_ERROR RPL_NO_PERM RPL_CRC_DIFF
"""
import functools
import logging
import copy

from stp_gui.Config import Config


# -----------------------------------------------------------------------------
# Create message
# -----------------------------------------------------------------------------

def chunks(l, n):
    """
    Yield successive n-sized chunks from l.
    :param l:
    :param n:
    """
    for i in range(0, len(l), n):
        yield l[i:i+n]


def create_message(type, data=None, answer=None, length=131072):
    """
    Create list with message-sequence. For simple message list contain one element
    :param type: message_type
    :param data: databody
    :param answer: message_answer
    :param length: max length for message (split message)
    """
    m_type = (lambda T: " " + T if T else "")
    m_answer = (lambda A: " " + A + "\n" if A else "\n")
    m_crc = (lambda C: "CRC: " + str(C) + "\n" if C else "")
    m_size = (lambda D: "Size: " + str(D) + "\n" if D else "")
    m_multi = (lambda M: "Multiple: " + str(M) + "\n" if M else "")
    m_number = (lambda M: "Multi-Number: " + str(M) + "\n" if M else "")
    g_message = (lambda t, a, s, c, d: str("STP" + t + a + s + c + "\n" + d))
    g_multi = (lambda t, a, s, c, m, n, d: str("STP" + t + a + s + c + m + n + "\n" + d))

    cfg_crc = Config().crc()

    messagelist = []
    if data:
        datalist = list(chunks(data, length))
        m = len(datalist)
        if m > 1:
            for i, d in enumerate(datalist):
                crc = cfg_crc
                messagelist.append(g_multi(m_type(type), m_answer(answer),
                                            m_size(len(d)), m_crc(crc),
                                            m_multi(m), m_number(i+1), d))
        else:
            crc = cfg_crc
            messagelist = [g_message(m_type(type), m_answer(answer),
                                      m_size(len(data)), m_crc(crc), data)]
    else:
        crc = 0
        messagelist = [g_message(m_type(type), m_answer(answer),
                                  m_size(None), m_crc(crc), "")]

    return messagelist


# -----------------------------------------------------------------------------
# Parse message
# -----------------------------------------------------------------------------

class InvalidProtocol(Exception):
    """ error raised when first line not contain 'STP'
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


# -----------------------------------------------------------------------------
class StpParser:
    """ Parse input message
    """

    def __init__(self, message=None):
        self.__messTail = ""  # tail symbols from prev message
        self.__reset()
        self.add_message(message)

    def __reset(self):
        self.errstr = ""
        self.__complete = False
        self.__messageDict = {}  # Keys: Type Answer Size CRC Multiple Multi-Number Body
        self.__tmpDict = {}
        self.__body = None  # databody in JSON format
        self.__multiData = {}

    # Input interface

    def add_message(self, data):
        """ Add received frame to parser
        """
        message = str(self.__messTail) + str(data)
        self.__messTail = ""
        if message[:3] != 'STP':
            self.__reset()
            return False
        else:
            self.__parse(message)
            return self.__complete

    # Private part

    def __parse(self, message):
        """ Parsed message from raw data
        """
        if not message:
            self.__reset()
            return
        separator = '\n'
        message = message.replace('\r\n', separator)
        # split to head-body
        msg_st = list(message.split(separator * 2))
        tHead = msg_st.pop(0)
        if msg_st:
            self.__body = functools.reduce(lambda a, b: str(a) + (separator * 2) + str(b), msg_st)
        else:
            # not complete message
            self.__reset()
            return
        head = tHead.split(separator)
        try:
            self.errstr = ""
            if not self.__parse_head(head):
                self.__reset()
                return
            if self.__tmpDict.get('Multi-Number', 1) == 1:
                self.__add_single()
            else:
                self.__add_multi()
            # parse tail / next message
            if self.__messTail:
                self.add_message("")
        except InvalidProtocol as e:
            self.__reset()
            self.errstr = str(e)
            self.__reset()
            # TODO: raise exception

    def __parse_head(self, head):
        """ Parsed data as list of message head
        """
        # start line: STP <NetCmd.*> <NetAns.*>
        startstr = head.pop(0)
        d = startstr.split(" ")
        protocol = d[0]
        if protocol != "STP":
            raise InvalidProtocol('First word must be "STP": %s' % startstr)
        if len(d) == 2:
            self.__tmpDict['Type'] = str(d[1])
            logging.debug("Request detect: %s", startstr)
        elif len(d) == 3:
            self.__tmpDict['Type'] = str(d[1])
            self.__tmpDict['Answer'] = str(d[2])
            if startstr != 'STP MES_STATUS RPL_OK':
                logging.debug("Answer detect: %s", startstr)
        else:
            raise InvalidProtocol('Found too much words in start-string: %s' % startstr)

        # Messages without body don't have size so we initialize it to 0
        self.__tmpDict['Size'] = 0

        # header (key: value):
        for h in head:
            k, v = h.split(':')
            k = k.strip()
            v = v.strip()
            try:
                self.__tmpDict[str(k)] = int(v)
            except ValueError:
                pass
        # Check params
        if self.__tmpDict.get('Multiple', 0) == 0:
            self.__complete = True
        size = self.__tmpDict.get('Size', 0)
        if not size:
            self.__body = ""
            self.__tmpDict['Body'] = ""
            self.__messTail = ""
        else:
            if len(self.__body) < size -1:
                # wait next part of mesage
                return False
            self.__tmpDict['Body'] = self.__body[:size]
            self.__messTail = self.__body[size:]
        return True

    def __add_single(self):
        """ If single-frame message, or first multi-frame
        """
        self.__messageDict = copy.copy(self.__tmpDict)
        # first multi-frame
        if self.__messageDict.get('Multiple', 0):
            self.__multiData[self.__messageDict.get('Multi-Number', 1)] = self.__body

    def __add_multi(self):
        """ If multiple-message
        """
        n = self.__tmpDict.get('Multi-Number', 1)
        if self.__messageDict.get('Type') == self.__tmpDict.get('Type') and \
           self.__messageDict.get('Answer') == self.__tmpDict.get('Answer') and \
           self.__messageDict.get('CRC') == self.__tmpDict.get('CRC') and \
           self.__messageDict.get('Multiple') == self.__tmpDict.get('Multiple') and \
           int(self.__messageDict.get('Multi-Number')) + 1 == int(n):
            # save part message
            self.__multiData[n] = self.__body
            self.__messageDict['Multi-Number'] = n
            if n == self.__messageDict.get('Multiple'):
                self.__body = ""
                size = 0
                for i in range(1, n+1):
                    size = size + len(self.__multiData[i])
                    self.__body = self.__body + self.__multiData[i]
                self.__messageDict['Body'] = self.__body
                self.__messageDict['Size'] = size
                self.__complete = True
        else:
            raise InvalidProtocol("Error on multi-message parser")


    # Return current message properties:

    def is_complete(self):
        return self.__complete

    def get_type(self):
        return self.__messageDict.get('Type')

    def get_answer(self):
        return self.__messageDict.get('Answer')

    def get_size(self):
        return self.__messageDict.get('Size', 0)

    def get_crc(self):
        return self.__messageDict.get('CRC', 0)

    def get_multi(self):
        return self.__messageDict.get('Multiple', 0)

    def get_multi_number(self):
        return self.__messageDict.get('Multi-Number', 0)

    def get_data(self):
        return self.__messageDict.get('Body')

    def get_message(self):
        return self.__messageDict

    # Debug:

    def print_message(self):
        """ debug function for print all fields
        """
        if self.get_type() == 'MES_STATUS':
            return
        print ("*******************************************************************************")
        print ("Message %s" % ("complete" if self.is_complete() else "non complete"))
        print ("Type:     %s" % self.get_type())
        print ("Answer:   %s" % self.get_answer())
        print ("Size:     %d" % self.get_size())
        print ("CRC:      0x%x" % self.get_crc())
        print ("Multiple: %d" % self.get_multi())
        print ("Multi-N:  %d" % self.get_multi_number())
        print (self.get_data())
        print ("*******************************************************************************")


# -----------------------------------------------------------------------------
# No __main__
