import logging
import os
import sys

from PyQt5 import QtGui, QtCore, QtWebKit
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication

import stp_gui.Common as Common
from stp_gui.ExtraLogging import ExtraLog
from QtWebKitWidgets import *

# from timeit import default_timer as timer

sys.path.append(os.path.join(Common.stp_dir, 'libpy/lib/svgwrite-1.1.9'))
sys.path.append(os.path.join(Common.stp_dir, 'libpy/lib/pyparsing-2.1.10'))

try:
    import svgwrite
    SVGWRITE_INSTALLED = True
except ImportError:
    SVGWRITE_INSTALLED = False


def _(text):
    return Common._(text)

class MonitoringWindow(ExtraLog, QWebView):

    # We will not create an instance of this class if files
    # scheme.svg, begin.end and end.html were not found
    def __new__(cls, tree):
        if not os.path.isfile(os.path.join(Common.data_dir, 'user-scheme.svg')):
            logging.info(_("SVG is absent. Trying to generate SVG from config."))
            if not MonitoringWindow.generateSVG(tree.mainWindow.stpCtrl.get_config()):
                return
        else:
            try:
                with open(os.path.join(Common.data_dir, 'user-scheme.svg'), 'r') as user_scheme, \
                        open(os.path.join(Common.data_dir, 'scheme.svg'), 'w') as pic:
                    pic.write(user_scheme.read())
            except IOError as e:
                logging.warning(_("Can't load file: {}".format(e.filename)))

        try:
            with open(os.path.join(Common.data_dir, 'scheme.svg'), 'r') as pic, \
                    open(os.path.join(Common.data_dir, 'begin.html'), 'r') as begin, \
                    open(os.path.join(Common.data_dir, 'end.html'), 'r') as end, \
                    open(os.path.join(Common.data_dir, 'scheme_result.html'), 'w') as res:
                lines = pic.readlines()
                for line in lines:
                    res.write(line.replace('</svg>', ''))

                res.write(begin.read())
                res.write(pic.read())
                res.write(end.read())
                # res.write('</svg>\n')

        except IOError as e:
            logging.warning(_("Can't load file: {}".format(e.filename)))
            return None

        return QWebView.__new__(cls)

    def __init__(self, tree):
        super(MonitoringWindow, self).__init__()
        self.debug = False
        self.tree = tree

        self.tree.statusUpdated.connect(self.update)
        self.tree.statusBeginUpdate.connect(self.beginUpdate)
        self.tree.statusEndUpdate.connect(self.endUpdate)
        self.tree.openMonitoring.connect(self.tree.openMonitorView)
        self.tree.closeMonitoring.connect(tree.closeMonitorView)
        self.settings().setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)

        self.setGeometry(100, 50, 800, 600)
        self.setWindowTitle('STP::Monitoring')

        self.jstring = ''

        self.setContextMenuPolicy(Qt.NoContextMenu)

    def show(self):
        self.loadFinished.connect(self.afterLoad)
        self.load(QtCore.QUrl(os.path.join(Common.data_dir, 'scheme_result.html')))
        if self.debug:
            self.setContextMenuPolicy(Qt.DefaultContextMenu)
            # self.inspect = QtWebKit.QWebInspector()
            # self.inspect.setPage(self.page())
            # self.inspect.setVisible(True)
        super(MonitoringWindow, self).show()
        self.tree.openMonitoring.emit()

    def closeMonitoring(self):
        self.close()

    def closeEvent(self, event):
        self.tree.closeMonitoring.emit()

    def beginUpdate(self):
        self.jstring = ''

    def endUpdate(self):
        self.page().mainFrame().evaluateJavaScript(self.jstring)

    '''
	 "system": {
		state: state,
 		loop: 0,
		time: seconds,
		nowrun: 0,
		passes: 0,
		skips: 0,
		errors: 0,
		id: id_UI_with_exclusive,
		key: "unique_session_key",
	},
	"hosts": [
		{
			"hostname": vkname,
			state: state,
			loop: 0,
			time: seconds,
			nowrun: 0,
			passes: 0,
			skips: 0,
			errors: 0,
			"tests": [
				{
					"testname": testname,
					loop: 0,
					time: seconds,
					nowrun: 0,
					passes: 0,
					skips: 0,
					errors: 0,
					state: state,
					0:[...],
					...
					i:[...]
				},
    '''

    def update(self, host):
        #        self.t = timer()
        text = ''

        total = host.get('loop', 0)
        time = host.get('time', 0)
        passes = host.get('passes', 0)
        errors = host.get('errors', 0)
        state = host.get('state', 0)
        #        nowrun = host.get('nowrun', 0)
        hostname = host['hostname']
        running = self.tree.mainWindow.stpCtrl.running

        text = 'total: {}\\ntime: {}\\nsuccess: {}\\nerrors: {}\\n'.format(total, time, passes, errors)

        # print '%s %d %d - %s' % (hostname, state, nowrun, str(running))
        # 0x0010 - STOP 0x40000 - ENABLE
        if not state & 0x40000:
            self.jstring += self.updateHostColour(hostname, 'grey')
        elif running:
            if not errors:
                self.jstring += self.updateHostColour(hostname, 'YellowGreen')
            else:
                self.jstring += self.updateHostColour(hostname, 'Orange')
        elif not errors and not passes and not time and not total:
            self.jstring += self.updateHostColour(hostname, 'LimeGreen')
        elif not errors and passes:
            self.jstring += self.updateHostColour(hostname, 'green')
        elif errors:
            self.jstring += self.updateHostColour(hostname, 'red')
        else:
            self.jstring += self.updateHostColour(hostname, 'yellow')

        self.jstring += self.updateResults(hostname, None, text)

        # generate for all test status
        for test in host.get('tests', ''):
            self.generateTestsString(hostname, test, running)

        #        print("{:10s} {:.10f} secs".format(hostname, timer() - self.t))
        return

    def generateTestsString(self, hostname, test, running):
        testname = test['testname']
        total = test.get('loop', 0)
        time = test.get('time', 0)
        passes = test.get('passes', 0)
        errors = test.get('errors', 0)
        state = test.get('state', 0)

        text = 'total: {}\\ntime: {}\\nsuccess: {}\\nerrors: {}\\n'.format(total, time, passes, errors)

        # print '%s %s : %d %d %d' % (hostname, testname, test['state'], test['nowrun'], test['time'])
        # 0x0010 - STOP 0x40000 - ENABLE
        if not state & 0x40000:
            self.jstring += self.updateTestColour(hostname, testname, 'grey')
        elif running:
            if not errors:
                self.jstring += self.updateTestColour(hostname, testname, 'Olive')
            else:
                self.jstring += self.updateTestColour(hostname, testname, 'OrangeRed')
        elif not errors and not passes and not time and not total:
            self.jstring += self.updateTestColour(hostname, testname, 'LimeGreen')
        elif not errors and passes:
            self.jstring += self.updateTestColour(hostname, testname, 'ForestGreen')
        elif errors and passes:
            self.jstring += self.updateTestColour(hostname, testname, 'OrangeRed')
        elif errors and not passes:
            self.jstring += self.updateTestColour(hostname, testname, 'FireBrick')
        else:
            self.jstring += self.updateTestColour(hostname, testname, 'Gold')

        self.jstring += self.updateResults(hostname, testname, text)

    def updateResults(self, hostname, testname, text):
        js = ''
        if testname is None:
            # update Text in SVG
            js += 'generateInfo("' + hostname + '", "' + text + '");'

            # update ToolTip
            js += 'changeInfo("' + hostname + '", "' + text + '");'

            # update modal window
            js += 'updateModal("' + hostname + '", "' + text + '");'
        else:
            test_id = hostname + '_' + testname
            # update modal window
            js += 'updateModal("' + test_id + '", "' + text + '");'

        return js

    def updateTestColour(self, hostname, testname, color):
        return 'updateTestColour("' + hostname + '", "' + testname + '", "' + color + '");'

    def updateHostColour(self, hostname, color):
        return 'updateHostColour("' + hostname + '", "' + color + '");'

    def afterLoad(self, success):
        if not success:
            logging.warning(_("Can't load file: data/scheme.svg."))
        else:
            root = self.tree.getNodeByPath('/')
            if not root:
                return
            children_num = root.childCount()
            hosts = [root.child(i).getPatch()[1:] for i in range(0, children_num)]
            js = 'addListeners({})'.format(hosts)
            self.page().mainFrame().evaluateJavaScript(js)
            js = 'addInfo({})'.format(hosts)
            self.page().mainFrame().evaluateJavaScript(js)

            for host in hosts:
                node = self.tree.getNodeByPath('/' + host)
                children_num = node.childCount()
                tests = [node.child(i).getPatch()[1:] for i in range(0, children_num)]
                ids = [x.replace('/', '_') for x in tests]
                js = 'addTestListeners({})'.format(ids)
                self.page().mainFrame().evaluateJavaScript(js)

    @staticmethod
    def generateSVG(cfg, checked=False, num_of_hosts=12):
        if not SVGWRITE_INSTALLED:
            logging.warning(_("Can't load svgwrite module. Autogeneration of SVG file failed"))
            return False
        screen_geometry = QApplication.desktop().screenGeometry()
        height = screen_geometry.height()
        width = screen_geometry.width()
        # cfg = Config()
        if not cfg:
            logging.warning(_("Can't generate SVG: No configuration loaded"))
            return
        hosts = [h.hostname for h in cfg.hosts]
        if not hosts:
            logging.warning(_("Can't generate SVG: No hosts present in configuration"))
            return
        num_of_hosts = len(hosts)

        dwg = svgwrite.Drawing(os.path.join(Common.data_dir, 'scheme.svg'))
        dwg.viewbox(0, 0, width, height)

        columns = 0
        rows = int(columns + round(float(width) / float(height)) - 1)

        while columns * rows < num_of_hosts:
            columns += 1
            rows = int(columns + round(float(width) / float(height)) - 1)

        sz = int(min(height, width)) / columns
        for x in range(0, rows):
            for y in range(0, columns):
                if hosts:
                    name = hosts.pop()
                    group = dwg.g(id=name)
                    rect = dwg.rect(insert=(x * sz, y * sz),
                                    size=(sz * 1 - 10, sz - 10),
                                    fill='lightgrey',
                                    stroke='black',
                                    stroke_width=3)

                    rect2 = dwg.rect(insert=(x * sz, y * sz),
                                     size=(sz * 1 - 10, (sz - 10) * .2),
                                     fill='green',
                                     stroke='black',
                                     stroke_width=3,
                                     id=name + 'Data')

                    text_x = x * sz + (sz - 10) / 2
                    text_y = y * sz + (sz - 10) * .2 / 2
                    text = dwg.text(name,
                                    insert=(text_x, text_y),
                                    font_size=sz / 10,
                                    text_anchor='middle')

                    text2_x = x * sz + sz / 2
                    text2_y = y * sz + sz / 4 + sz / 16
                    text2 = dwg.text(name,
                                     insert=(text2_x, text2_y),
                                     font_size=sz / 10,
                                     id=name + 'Text',
                                     text_anchor='middle')
                    group.add(rect)
                    group.add(rect2)
                    group.add(text)
                    group.add(text2)
                    dwg.add(group)
                else:
                    rect = dwg.rect(insert=(x * sz, y * sz), size=(sz, sz), fill='lightgrey', visibility='hidden')
                    dwg.add(rect)

        dwg.save()
        return True
