#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Stp Controller

Contains current config, config file name and editor status.
Establish connect to kernel and send messages from kernel to UI.
Receive commands from UI and translate it to kernel.


FSM:
(old_state) --transition--> (new_state)
  event
  \ condition
  => action


Start --start app--> (START)


(START) --connect--> (CONNECTED)
  NetConnect - GUI: button "Connect" pressed or key in commandline
  => __tr_connect() - try connect to kernel, send MES_ACKNOWLEDGE

(START) --fopen--> (EDITOR)
  ConfigCreate or ConfigRead - GUI: config file opened or created or key in commandline
  => __cfg.create() or __cfg.read()


(CONNECTED) --net_indication--> (CONNECTED)
  NetAckInd - MES_ACKNOWLEDGE RPL_OK
  => request config from kernel (GET_CFG)

(CONNECTED) --disconnect--> (START)
  NetDisconnect - NET: lost connection, GUI: button "Disconnect" pressed
  \ no config
  => __tr_disconnect() - disconnect from kernel

(CONNECTED) --disconnect--> (EDITOR)
  NetDisconnect - NET: lost connection, GUI: button "Disconnect" pressed
  \ config present
  => __tr_disconnect() - disconnect from kernel

(CONNECTED) --netconfig--> (READY)
  NetNewConfig - NET: coonfig received from kernel
  \ Self and kernel configs are equal or empty or user apply kernel config
  => __tr_netconfig() - compare configs and send question to UI

(CONNECTED) --netconfig--> (SYNCONFIG)
  NetNewConfig - NET: coonfig received from kernel
  \ Configs not equal
  => __tr_netconfig() - send self config to kernel


(EDITOR) --fclose--> (START)
  ConfigClose - GUI: file closed by user

(EDITOR) --connect--> (CONNECTED)
  NetConnect - GUI: button "Connect" pressed
  => __tr_connect() - try connect to kernel, send MES_ACKNOWLEDGE

(EDITOR) --fopen--> (EDITOR)
  ConfigCreate or ConfigRead


(SYNCONFIG) --kernelconfig--> (READY)
  ConfigFromKer
  => __tr_kernelconfig() - apply config from kernel

(SYNCONFIG) --disconnect--> (EDITOR)
  NetDisconnect - NET: lost connection, GUI: stay with self config
  => __tr_disconnect() - disconnect from kernel


(READY) --disconnect--> (EDITOR)
  NetDisconnect - NET: lost connection, GUI: button "Disconnect" pressed
  => __tr_disconnect() - disconnect from kernel

(READY) --fopen--> (READY)
  ConfigCreate or ConfigRead - GUI: open or create file
  \ Config not applyed by kernel
  => __tr_netconfig() - send self config to kernel

(READY) --fopen--> (SYNCONFIG)
  ConfigCreate or ConfigRead - GUI: open or create file
  \ Config applyed by kernel
  => __tr_netconfig() - send self config to kernel

(READY) --run--> (RUNNING)
  CmdRun - GUI: button "Run" pressed
  => __tr_run(True) - Start tests


(RUNNING) --disconnect--> (EDITOR)
  NET: потеря соединения
  GUI: нажата кнопка "Disonnect"
  => разорвать соединение с ядром

(RUNNING) --stop--> (READY)
  GUI: нажата кнопка "Stop"
  => останов тестирования

(RUNNING) --pause--> (PAUSED)
  GUI: нажата кнопка "Pause"
  => приостановить тестирование


(PAUSED) --disconnect--> (EDITOR)
  NET: потеря соединения
  GUI: нажата кнопка "Disonnect"
  => разорвать соединение с ядром

(PAUSED) --stop--> (READY)
  GUI: нажата кнопка "Stop"
  => останов тестирования

(PAUSED) --continue--> (RUNNING)
  GUI: нажата кнопка "Run"
  => возобновление тестирования


(START) --выход--> Завершение программы
  GUI: нажата кнопка "Exit"

"""
import logging
import json
from PyQt5 import QtCore
from PyQt5.QtCore import QObject
from stp_gui.Common import APP_VERSION
from stp_gui.Config import Config
from stp_gui.Config import host, test
from stp_gui.StpNetwork import StpNetwork, NetCmd, NetAns
from binascii import crc32 as sum32

from stp_gui.Common import _

# -----------------------------------------------------------------------------
class State(object):
    """ FSM state: """
    START = 'START'  # start app (no connect and no config)
    CONNECTED = 'CONNECTED'  # connect to kernel established
    EDITOR = 'EDITOR'  # config editor only (co connect)
    SYNCONFIG = 'SYNCONFIG'  # sync local config with kernel
    READY = 'READY'  # test finished (or not running)
    RUNNING = 'RUNNING'  # test running
    PAUSED = 'PAUSED'  # test stopped


class FsmEvent(object):
    """ Events from GUI & Net: """
    Unknown = 'Unknown'
    ConfigCreate = 'ConfigCreate'  # create empty config
    ConfigRead = 'ConfigRead'  # open config file
    ConfigClose = 'ConfigClose'  # close config
    NetConnect = 'NetConnect'  # connect to kernel
    NetDisconnect = 'NetDisconnect'  # disconnect from kernel
    NetNewConfig = 'NetNewConfig'  # config from kernel received
    NetAckInd = 'NetAckInd'  # MES_ACKNOWLEDGE RPL_OK received
    ConfigFromKer = 'ConfigFromKer'  # apply config received from kernel
    # ConfigSelf    = 'ConfigSelf'    # use self (current) config
    # ConfigOk      = 'ConfigOk'      # self config applyed by kernel
    CmdRun = 'CmdRun'  # start testing
    CmdPause = 'CmdPause'  # pause testing
    CmdStop = 'CmdStop'  # stop testing
    CmdExit = 'CmdExit'  # exit app


# -----------------------------------------------------------------------------
class StpControl(QObject):
    """ StpControl class """

    # emit signal to GUI:
    changeState = QtCore.pyqtSignal()  # if changed: editmode, testmode, running
    diffConfig = QtCore.pyqtSignal(str, str)  # new config or MDF_* received from kernel
    statsUpdate = QtCore.pyqtSignal()  # CMD_* received from kernel
    heartBeat = QtCore.pyqtSignal(str)  # MES_STATUS received from kernel
    exclusive_signal = QtCore.pyqtSignal(int)
    logReceived = QtCore.pyqtSignal(str)
    infoReceived = QtCore.pyqtSignal(str)

    # singleton
    def __new__(cls, *args, **kw):
        if not hasattr(cls, '_instance'):
            cls._instance = super(StpControl, cls).__new__(cls, *args, **kw)
        return cls._instance

    def __init__(self, parent):
        super().__init__()
        self.__parent = parent
        self.__cfg = Config()
        self.__cfg.add_observer(self)
        self.__netcfgjson = None
        self.__netcfgcrc = 0
        self._client = StpNetwork(self)
        self._client.sigIndication.connect(self.net_indication)
        self._client.sigError.connect(self.net_error)
        # FSM state
        self.__state = State.START
        # Status for GUI:
        self.__validconfig = False
        self._connected = False
        self.__exclusive = False
        self.__autostart = False
        self.__session = False
        self.__english = False
        self.__editmode = True
        self.__testmode = False
        self.__running = False
        self.__paused = False

        self.editor_callbacks = []

        self.__stp_id = -1
        self.__heartbeat_ctr = 0
        # attention:
        # when no_skip is False - we skipped net messages about status
        # skip_step - set skip count
        self.no_skip = True
        self.skip_step = 2

    # ------------------------------------
    #           STATUS
    # ------------------------------------

    @property
    def stp_id(self):
        return self.__stp_id

    @property
    def validconfig(self):
        """ config file opened and equal to kernel config """
        return self.__validconfig

    @property
    def connected(self):
        """ connect to kernel established """
        return self._connected

    @property
    def exclusive(self):
        """ connected to kernel in exclusive mode """
        return self.__exclusive

    @property
    def autostart(self):
        """ send run cmd to at start """
        return self.__autostart

    @property
    def paused(self):
        """ in PAUSED state """
        return self.__paused

    @property
    def english(self):
        """ setup english language """
        return self.__english

    @property
    def session(self):
        """ load session, cfg will be ignored """
        return self.__session

    @property
    def editmode(self):
        """ config file edit enabled """
        return self.__editmode

    @property
    def testmode(self):
        """ test run/stop enabled """
        return self.__testmode

    @property
    def running(self):
        """ test runninf """
        return self.__running

    @property
    def state(self):
        return self.__state

    def __fsm_update(self, nstate):
        """ update properties and switch state for FSM """
        if self.__state == nstate:
            return
        logging.debug("State changed: %s -> %s" % (str(self.__state), str(nstate)))
        self.__state = nstate
        self.__paused = False
        if self.__state == State.START or \
                self.__state == State.CONNECTED or \
                self.__state == State.EDITOR:
            # test run/stop disabled
            self.__editmode = True
            self.__testmode = False
            self.__running = False
            for f in self.editor_callbacks:
                f()
        elif self.__state == State.RUNNING:
            # config file edit edisabled
            self.__editmode = False
            self.__testmode = False
            self.__running = True
        elif self.__state == State.PAUSED:
            # config file edit edisabled
            self.__editmode = False
            self.__testmode = False
            self.__running = False
            self.__paused = True
        elif self.__state == State.READY:
            # all enabled
            self.__editmode = True
            self.__testmode = True
            self.__running = False
        elif self.__state == State.SYNCONFIG:
            # all enabled
            self.__editmode = False
            self.__testmode = False
            self.__running = False
        # send event to User
        self.changeState.emit()

    # ------------------------------------
    #           CONFIG
    # ------------------------------------

    def config_ischanged(self):
        """ check config changed """
        return self.__cfg.is_changed()

    def config_getfilename(self):
        """ get current filename for config """
        return self.__cfg.get_filename()

    def config_create(self):
        """ create empty config """
        self.__run_fsm(FsmEvent.ConfigCreate)

    def config_read(self, filename):
        """
        read config from file
        :param filename:
        """
        self.__run_fsm(FsmEvent.ConfigRead, str(filename))

    def config_write(self, filename=None):
        """
        save config to (current) file
        :param filename:
        """

        if filename is not None:
            logging.debug('Saving config. filename: ' + str(filename))
        else:
            logging.debug('Saving changes in config.')
        return self.__cfg.write(filename)

    def config_present(self):
        # TODO: config_present should be moved to Cinfig class since checking
        # Config.closed variable there is more logical
        """ check present or none config """
        if self.__cfg.get_cfg() and not Config.closed:
            return True
        else:
            return False

    def get_config(self):
        """ return current config """
        return self.__cfg.get_cfg()

    def get_schema_parser(self):
        """ return config schema """
        return self.__cfg.get_schema_parser()

    # ------------------------------------
    #           NETWORK
    # ------------------------------------

    def stp_connect(self, ipaddr="localhost", port="9876"):
        """
        try connect to kernel
        :param ipaddr:
        :param port:
        """
        self.__run_fsm(FsmEvent.NetConnect, ipaddr, port)

    def stp_disconnect(self, kernclose=False, monclose=False):
        """
        disconnect from kernel
        :param kernclose:
        :param monclose:
        """
        self.__run_fsm(FsmEvent.NetDisconnect, kernclose, monclose)

    def stp_setexcl(self, excl=True):
        """
        request to set/clean exclusive mode
        :param excl:
        """
        logging.debug("Set exclusive")
        if self.__exclusive == excl:
            return
        mode = None
        if excl:
            mode = NetCmd.SET_EXCL
            # TODO: update on responce
            # self.__exclusive = True
        else:
            mode = NetCmd.SET_UN_EXCL
            # TODO: update on responce
            # self.__exclusive = False
        self._client.net_control(mode)

    def stp_setautostart(self, flag=True):
        """
        request to set/clean autostart mode
        :param autostart:
        """
        if flag:
            logging.debug("Set autostart")
        else:
            logging.debug("Clear autostart")
        self.__autostart = flag

    def stp_reconnectmon(self):
        """ reconnect to all stp_monitor """
        self._client.net_control(NetCmd.WRK_RECONNECT)

    # ------------------------------------

    # TODO: update on responce
    # run self._client.net_control(NetCmd.
    # fsm run from indication

    def test_start(self):
        """ test run/resume """
        logging.debug("Start TEST")
        # TODO: call self.__tr_run() here and __run_fsm in net_indication
        self.__tr_run(True)
        # self.__run_fsm(FsmEvent.CmdRun)

    def test_pause(self):
        """ test pause """
        logging.debug("Pause TEST")
        self.__run_fsm(FsmEvent.CmdPause)

    def test_stop(self):
        """ test stop """
        logging.debug("Stop TEST")
        self.__run_fsm(FsmEvent.CmdStop)

    def test_reset(self):
        """ clean test result """
        logging.debug("Reset TEST")
        self._client.net_control(NetCmd.CMD_RESET)

    # ------------------------------------

    def running_to_ready(self):
        """
        Explicitly set from RUNNING to READY
        """
        logging.debug("Testing finished")
        self.__tr_stop()
        self.__fsm_update(State.READY)

    def to_running(self):
        """
        Explicitly set RUNNING state
        """
        if self.__state != State.RUNNING:
            self.__fsm_update(State.RUNNING)

    def to_paused(self):
        """
        Explicitly set PAUSED state
        """
        if self.__state != State.PAUSED:
            self.__fsm_update(State.PAUSED)

    def to_ready(self):
        """
        Explicitly set READY state
        """
        if self.__state == State.CONNECTED:
            # Going from CONNECTED to READY will break config synchronization
            return
        if self.__state != State.READY:
            self.__fsm_update(State.READY)

    # ------------------------------------

    def model_changed(self):
        """
        Model observer
        Called when model is changed
        """
        logging.debug("model_changed")

    # ------------------------------------

    def get_log(self, host, test, instance):
        logging.debug("Request test log")
        data = '{"instance":' + str(instance) + ',"testname":"' + test + '","vkname":"' + host + '"}'
        self._client.net_request(NetCmd.GET_TEST_LOG, data)

    def get_info(self, host):
        logging.debug("Request host info")
        data = '{"vkname":"' + host + '"}'
        self._client.net_request(NetCmd.GET_INFO_ABOUT_VK, data)

    def send_cfg(self, new_cfg=False):
        if new_cfg:
            # TODO: It would be better to create an instance of Config and dump
            # it immediately, but Config is a singleton. Most likely we don't
            # need it to be singleton, so we must try to refactor it later
            empty_cfg = '{"options":{"autostart":0,"concurrency":0,"enable":1,"errors":1,"max_sys_errors":0,"max_sys_passes":0,"max_sys_time":0,"passes":0,"realtime":0,"reconnect_time":0,"repeat":1,"singlepass":0,"time":0}}'
            self._client.net_request(NetCmd.SET_CFG, empty_cfg)
        # print 'cfg:', self.__cfg.dump(compact=True, remdef=True)
        else:
            self._client.net_request(NetCmd.SET_CFG, self.__cfg.dump(compact=True, remdef=True))

    def close_cfg(self):
        if self.connected:
            self._client.net_request(NetCmd.SET_CFG)
        else:
            Config.closed = True
            cfg = Config()
            cfg.create()
            self.diffConfig.emit(NetCmd.SET_CFG, self.__cfg.dump(compact=True, remdef=True))

    # ------------------------------------
    #         Network Indication
    # ------------------------------------

    def net_error(self, errstr):
        """ any error on network
        """
        logging.error(_("Network error: ") + errstr)
        if self.__state != State.START:
            self.__run_fsm(FsmEvent.NetDisconnect, False, False)

    def net_indication(self, mes):
        """ received message from kernel
        """
        # FIXME:
        mtype = mes.get_type()
        mans = mes.get_answer()
        if not mes.get_data() and mtype == NetCmd.SET_CFG:
            if mans == NetAns.RPL_NO_PERM:
                return
            self.__netcfgjson = str(mes.get_data())
            self.__netcfgcrc = sum32(self.__netcfgjson.encode('utf-8'))
            logging.info(_("Configuration closed"))
            Config.closed = True
            cfg = Config()
            cfg.create()
            self.diffConfig.emit(NetCmd.SET_CFG, self.__cfg.dump(compact=True, remdef=True))
            self.__fsm_update(State.CONNECTED)
            return
        # on connect:
        if mtype == NetCmd.MES_ACKNOWLEDGE:
            if mans == NetAns.RPL_OK:
                s = mes.get_data().replace("'", '"')
                self.__stp_id = json.loads(s)['id']
                self.__run_fsm(FsmEvent.NetAckInd)
            else:
                logging.debug("Network event")
                self.__run_fsm(FsmEvent.NetDisconnect, False, False)
        # heartbeat:
        elif mtype == 'MES_STATUS':
            # if mans == NetAns.RPL_OK:
            if self.no_skip:
                self.__heartbeat_ctr = self.skip_step
            if mans == NetAns.RPL_OK and self.__heartbeat_ctr == self.skip_step:
                self.heartBeat.emit(mes.get_data())
                self.__heartbeat_ctr = -1
            self.__heartbeat_ctr += 1
        # config from kernel:
        elif mtype == NetCmd.GET_CFG:
            if mans == NetAns.RPL_NONE:
                logging.info(_("No config in kernel"))
                self.__run_fsm(FsmEvent.NetNewConfig, True)
            elif mans == NetAns.RPL_OK:
                logging.info(_("Receive config from kernel"))
                self.__netcfgjson = str(mes.get_data())
                self.__netcfgcrc = sum32(self.__netcfgjson.encode('utf-8'))
                self.__run_fsm(FsmEvent.NetNewConfig, False)
        # self config ack from kernel:
        elif mtype == NetCmd.SET_CFG:
            if mans == NetAns.RPL_OK:
                Config.closed = False
                logging.info(_("Config applyed by kernel"))
                self.__netcfgjson = str(mes.get_data())
                self.__netcfgcrc = sum32(self.__netcfgjson.encode('utf-8'))
                self.__run_fsm(FsmEvent.ConfigFromKer)
            elif mans == NetAns.RPL_NO_PERM:
                self._client.net_request(NetCmd.GET_CFG)
                logging.error(_("RPL_NO_PERM on send config"))
                # self.__run_fsm(FsmEvent.NetDisconnect, False, False)

        elif mtype == NetCmd.CMD_RUN:
            if mans == NetAns.RPL_OK:
                self.__run_fsm(FsmEvent.CmdRun)
            else:
                self.diffConfig.emit(mtype, '')
        elif mtype == NetCmd.CMD_PAUSE or \
                mtype == NetCmd.CMD_STOP:
            if mans != NetAns.RPL_OK:
                self.diffConfig.emit(mtype, '')

        # CMD_ answers
        # NetCmd.CMD_RUN
        # NetCmd.CMD_PAUSE
        # NetCmd.CMD_CONTINUE
        # NetCmd.CMD_STOP
        # NetCmd.CMD_RESET
        # self.statsUpdate.emit()
        # MDF_ answers:
        elif mtype == NetCmd.MDF_ADD_VK or \
                mtype == NetCmd.MDF_CHN_VK or \
                mtype == NetCmd.MDF_DEL_VK or \
                mtype == NetCmd.MDF_ADD_TEST or \
                mtype == NetCmd.MDF_CHN_TEST or \
                mtype == NetCmd.MDF_DEL_TEST or \
                mtype == NetCmd.MDF_CHN_PREF or \
                mtype == NetCmd.MDF_VK_ON or \
                mtype == NetCmd.MDF_VK_OFF or \
                mtype == NetCmd.MDF_VKALL_INVERT or \
                mtype == NetCmd.MDF_VKALL_ON or \
                mtype == NetCmd.MDF_VKALL_OFF or \
                mtype == NetCmd.MDF_TEST_ON or \
                mtype == NetCmd.MDF_TEST_OFF or \
                mtype == NetCmd.MDF_TESTALL_INVERT or \
                mtype == NetCmd.MDF_TESTALL_ON or \
                mtype == NetCmd.MDF_TESTALL_OFF:
            if mans == NetAns.RPL_OK:
                Config.closed = False
                self.__fsm_update(State.READY)
                self.diffConfig.emit(mtype, mes.get_data())
            else:
                # TODO: error to user
                pass
        # NetCmd.MES_UI_DONE
        elif mtype == NetCmd.MES_MESSAGE:
            self.diffConfig.emit(mtype, mes.get_data())
        elif mtype == NetCmd.SET_EXCL:
            if mes.get_answer() == 'RPL_NO_PERM':
                logging.warning(_("Can't set exclusive mode"))
                return
            if not self.exclusive:
                self.__exclusive = True
                self.exclusive_signal.emit(True)
        elif mtype == NetCmd.SET_UN_EXCL:
            if self.exclusive:
                self.__exclusive = False
                self.exclusive_signal.emit(False)
        elif mtype == NetCmd.GET_TEST_LOG:
            data = mes.get_data()
            self.logReceived.emit(data)
        elif mtype == NetCmd.GET_INFO_ABOUT_VK:
            data = mes.get_data()
            self.infoReceived.emit(data)
        # NetCmd.MES_SERVER_DONE
        # NetCmd.WRK_TERMINATE
        # NetCmd.WRK_DISCONNECT
        # NetCmd.WRK_RECONNECT

    # ------------------------------------
    #           FSM
    # ------------------------------------

    def __tr_connect(self, ipaddr, port):
        """
        try connect to kernel, send MES_ACKNOWLEDGE
        :param ipaddr:
        :param port:
        """
        if self._client.net_connect(ipaddr, int(port)):
            try:
                v1, v2, v3 = APP_VERSION.split('.')
                avs = str("%d%03d%03d" % (int(v1), int(v2), int(v3)))
            except:
                avs = "9999999"
            self._client.net_request(NetCmd.MES_ACKNOWLEDGE, avs)
            self._connected = True
            self.__stp_id = -1
            return True
        return False

    def __tr_disconnect(self, kernclose, monclose):
        """
        disconnect from kernel
        :param kernclose:
        :param monclose:
        """
        mode = NetCmd.MES_UI_DONE
        if kernclose:
            mode = NetCmd.WRK_TERMINATE
        if monclose:
            mode = NetCmd.WRK_DISCONNECT

        if kernclose and monclose:
            mode = NetCmd.WRK_TERMALL

        self._client.net_control(mode)
        self._client.net_disconnect()
        self._connected = False
        self.__stp_id = -1
        self.__exclusive = False
        self.stp_setautostart(False)
        self.exclusive_signal.emit(False)

    def __run_autostart(self, s):
        if self.__autostart:
            logging.debug('Attempt RUN autostart: ' + str(s))
            self.__tr_run(True)
            self.stp_setautostart(False)

    def __tr_netconfig(self, emptycfg=False):
        """
        compare configs and send question to UI
        :param emptycfg: config in kernel empty
        return True if configs equal
        """
        if emptycfg:
            if self.config_present():
                logging.debug("Send self config to kernel")
                self.__tr_selfconfig()
                return False
            else:
                logging.debug("Kernel and self configs is empty")
                return True
        # received config from kernel

        if self.__netcfgcrc == self.__cfg.crc():
            if Config.closed:
                logging.debug("Self config is closed")
                Config.closed = False
                self.__tr_kernelconfig()
                return True
            else:
                logging.debug("Kernel config == self config")
                return True
        else:
            logging.debug("Kernel config != self config " + str(self.__netcfgcrc) + "||" + str(self.__cfg.crc()))
            if not self.__cfg.crc():
                self.__tr_kernelconfig()
                return True
            elif not self.cfg_changes_alowed() or self.__parent.selectConfig():
                self.__tr_kernelconfig()
                return True
            else:
                logging.debug("Send self config to kernel")
                self.__tr_selfconfig()
                return True

    def cfg_changes_alowed(self):
        if self.exclusive:
            # print 'excl'
            return True  # We are in exclusive mode
        elif self.__parent._excl_id == -1 or self.stp_id == -1:
            # print 'non excl'
            return True  # Nobody has exclusive mode available
        elif self.__parent._excl_id != self.stp_id:
            # print 'ro'
            return False  # We are in read only mode

    def __tr_kernelconfig(self):
        """ apply config received from kernel
        """
        logging.debug("Apply config from kernel")
        self.diffConfig.emit(NetCmd.SET_CFG, self.__netcfgjson)

    def __tr_selfconfig(self):
        """ send self config to kernel
        """
        logging.debug("Send self config to kernel")
        self._client.net_request(NetCmd.SET_CFG, self.__cfg.dump(compact=True, remdef=True))

    def __tr_run(self, mode):
        """запуск тестирования"""
        if mode:
            # запуск тестирования
            self._client.net_control(NetCmd.CMD_RUN)
        else:
            # возобновление тестирования
            self._client.net_control(NetCmd.CMD_CONTINUE)

    def __tr_stop(self):
        """останов тестирования"""
        self._client.net_control(NetCmd.CMD_STOP)

    def __tr_pause(self):
        """приостановить тестирование"""
        self._client.net_control(NetCmd.CMD_PAUSE)

    # MAIN FSM
    def __run_fsm(self, fsmevent, param1=None, param2=None):
        """ Stp Controller FSM
        """
        logging.debug("run_fsm: state=%s event=%s"
                      % (str(self.__state), str(fsmevent)))
        # State: START
        if self.__state == State.START:
            # (START) --connect--> (CONNECTED)
            if fsmevent == FsmEvent.NetConnect:
                if self.__tr_connect(param1, param2):
                    self.__fsm_update(State.CONNECTED)
            # (START) --fopen--> (EDITOR)
            elif fsmevent == FsmEvent.ConfigCreate:
                if self.__cfg.create():
                    self.__fsm_update(State.EDITOR)
            elif fsmevent == FsmEvent.ConfigRead:
                if self.__cfg.read(param1):
                    self.__fsm_update(State.EDITOR)
            else:
                logging.error(_("Unknown event {0} on state {1}").format(str(fsmevent), str(self.__state)))
        # State: CONNECTED
        elif self.__state == State.CONNECTED:
            # (CONNECTED) --net_indication--> (CONNECTED)
            if fsmevent == FsmEvent.NetAckInd:
                self._client.net_request(NetCmd.GET_CFG)
                # stay on CONNECTED
            # (CONNECTED) --disconnect--> (START/EDITOR)
            elif fsmevent == FsmEvent.NetDisconnect:
                self.__tr_disconnect(param1, param2)
                if self.config_present():
                    self.__fsm_update(State.EDITOR)
                else:
                    self.__fsm_update(State.START)
            # (CONNECTED) --netconfig--> (SYNCONFIG/READY)
            elif fsmevent == FsmEvent.NetNewConfig:
                logging.debug('net new config')
                if self.__tr_netconfig(param1):
                    if not self.config_present():
                        # auto create config are off
                        # self.config_create()
                        self.__tr_selfconfig()
                    # validconfig
                    else:
                        self.__fsm_update(State.READY)
                        self.__run_autostart(1)
                else:
                    # wait for kernel config
                    self.__fsm_update(State.SYNCONFIG)
                    logging.debug('SYNCONFIG')
            elif fsmevent == FsmEvent.ConfigCreate:
                # self.__cfg.create()
                self.__fsm_update(State.READY)
                self.send_cfg(new_cfg=True)
                # self.diffConfig.emit(NetCmd.SET_CFG, self.__cfg.dump(compact=True, remdef=True))
            elif fsmevent == FsmEvent.ConfigFromKer:
                logging.debug('config from ker')
                self.__tr_kernelconfig()
                self.__fsm_update(State.READY)
            elif fsmevent == FsmEvent.ConfigRead:
                if self.__cfg.read(param1):
                    if self.__tr_netconfig():
                        self.__fsm_update(State.READY)
                        self.stp_setautostart(False)
                    else:
                        self.__fsm_update(State.SYNCONFIG)
            else:
                logging.error(_("Unknown event {0} on state {1}").format(str(fsmevent), str(self.__state)))

        # State: EDITOR
        elif self.__state == State.EDITOR:
            # (EDITOR) --fclose--> (START)
            if fsmevent == FsmEvent.ConfigClose:
                self.__cfg = Config.Config()
                self.__fsm_update(State.START)
            # (EDITOR) --connect--> (CONNECTED)
            elif fsmevent == FsmEvent.NetConnect:
                if self.__tr_connect(param1, param2):
                    self.__fsm_update(State.CONNECTED)
            # (EDITOR) --fopen--> (EDITOR)
            elif fsmevent == FsmEvent.ConfigCreate:
                if self.__cfg.create():
                    Config.closed = False
                    self.__fsm_update(State.EDITOR)
            elif fsmevent == FsmEvent.ConfigRead:
                if self.__cfg.read(param1):
                    self.__fsm_update(State.EDITOR)
            elif fsmevent == FsmEvent.NetDisconnect:
                # We ignore NetDisconnect event since we are already disconnected
                pass
            else:
                logging.error(_("Unknown event {0} on state {1}").format(str(fsmevent), str(self.__state)))
        # State: SYNCONFIG
        elif self.__state == State.SYNCONFIG:
            logging.debug('SYN config')
            # (SYNCONFIG) --kernelconfig--> (READY)
            if fsmevent == FsmEvent.ConfigFromKer:
                self.__tr_kernelconfig()
                self.__fsm_update(State.READY)
            ######################################
            # this logic moved to __tr_netconfig()
            ## (SYNCONFIG) --selfconfig--> (SYNCONFIG)
            # elif fsmevent == FsmEvent.ConfigSelf:
            #    self.__tr_selfconfig()
            #    self.__fsm_update(State.SYNCONFIG)
            ## (SYNCONFIG) --selfconfig--> (READY)
            # elif fsmevent == FsmEvent.ConfigOk:
            #    self.__fsm_update(State.READY)
            ######################################
            # (SYNCONFIG) --disconnect--> (EDITOR)
            elif fsmevent == FsmEvent.NetDisconnect:
                self.__tr_disconnect(param1, param2)
                self.__fsm_update(State.EDITOR)
            else:
                logging.error(_("Unknown event {0} on state {1}").format(str(fsmevent), str(self.__state)))
        # State: READY
        elif self.__state == State.READY:
            # (READY) --disconnect--> (EDITOR)
            if fsmevent == FsmEvent.NetDisconnect:
                self.__tr_disconnect(param1, param2)
                self.__fsm_update(State.EDITOR)
            # (READY) --fopen--> (READY/SYNCONFIG)
            elif fsmevent == FsmEvent.ConfigCreate:
                # if self.__cfg.create():
                self.send_cfg(new_cfg=True)
                self.__cfg.create()
                self.__fsm_update(State.READY)
                self.stp_setautostart(False)
            elif fsmevent == FsmEvent.ConfigRead:
                if self.__cfg.read(param1):
                    if self.__tr_netconfig():
                        self.__fsm_update(State.READY)
                        self.stp_setautostart(False)
                    else:
                        self.__fsm_update(State.SYNCONFIG)
            # (READY) --run--> (RUNNING)
            elif fsmevent == FsmEvent.CmdRun:
                # self.__tr_run(True)
                self.__fsm_update(State.RUNNING)
            elif fsmevent == FsmEvent.ConfigFromKer:
                if self.__tr_netconfig():
                    self.diffConfig.emit(NetCmd.SET_CFG, self.__cfg.dump(compact=True, remdef=True))
                    self.__fsm_update(State.READY)
                    cfg = self.__cfg.get_cfg()
                    if cfg.hosts:
                        self.__run_autostart(2)
                    else:
                        self.stp_setautostart(False)
            elif fsmevent == FsmEvent.NetNewConfig:
                # If we are here, it means we received NO_PERM when sending cfg
                # to kernel earlier, so we should accept kernel's cfg
                self.diffConfig.emit(NetCmd.SET_CFG, self.__netcfgjson)
                self.stp_setautostart(False)
            else:
                logging.error(_("Unknown event {0} on state {1}").format(str(fsmevent), str(self.__state)))
        # State: RUNNING
        elif self.__state == State.RUNNING:
            # (RUNNING) --disconnect--> (EDITOR)
            if fsmevent == FsmEvent.NetDisconnect:
                self.__tr_disconnect(param1, param2)
                self.__fsm_update(State.EDITOR)
            # (RUNNING) --stop--> (READY)
            elif fsmevent == FsmEvent.CmdStop:
                self.__tr_stop()
                self.__fsm_update(State.READY)
            # (RUNNING) --pause--> (PAUSED)
            elif fsmevent == FsmEvent.CmdPause:
                self.__tr_pause()
                self.__fsm_update(State.PAUSED)
            else:
                logging.error(_("Unknown event {0} on state {1}").format(str(fsmevent), str(self.__state)))
        # State: PAUSED
        elif self.__state == State.PAUSED:
            # (PAUSED) --disconnect--> (EDITOR)
            if fsmevent == FsmEvent.NetDisconnect:
                self.__tr_disconnect(param1, param2)
                self.__fsm_update(State.EDITOR)
            # (PAUSED) --stop--> (READY)
            elif fsmevent == FsmEvent.CmdStop:
                self.__tr_stop()
                self.__fsm_update(State.READY)
            # (PAUSED) --continue--> (RUNNING)
            elif fsmevent == FsmEvent.CmdRun:
                self.__tr_run(False)
                self.__fsm_update(State.RUNNING)
            else:
                logging.error(_("Unknown event {0} on state {1}").format(str(fsmevent), str(self.__state)))
        else:
            logging.error(_("Unknown state %s") % (str(self.__state)))

    #  ////////////////////////////////////////////////////////////////////////////// StpControlMDF

    def mdf_addvk(self, name, dump=False):
        """Добавить новый ВК в текущую конфигурацию"""
        logging.debug("Add VK: %s" % name)
        # MDF_ADD_VK

        if dump:
            h = host(json.loads(dump))
            message = '{"vkname":"' + name + '","vk":[' + h.dump(compact=True) + ']}'
            if self._connected:
                self._client.net_request(NetCmd.MDF_ADD_VK, message)
            else:
                self.diffConfig.emit(NetCmd.MDF_ADD_VK, message)
        else:
            h = host()
            h.hostname = name
            message = '{"vkname":"' + name + '","vk":[' + h.dump(compact=True) + ']}'
            if self._connected:
                self._client.net_request(NetCmd.MDF_ADD_VK, message)
            else:
                self.diffConfig.emit(NetCmd.MDF_ADD_VK, message)
        return True

    def mdf_chnvk(self, hostname, dump):
        """Изменить параметры указанного ВК"""
        """in dump parameter - dump of Config.host_options class"""
        logging.debug("Change VK: %s" % hostname)
        # MDF_CHN_VK

        h = host(json.loads(dump))
        message = '{"vkname":"' + hostname + '","vk":[' + h.dump(compact=True) + ']}'
        if self._connected:
            self._client.net_request(NetCmd.MDF_CHN_VK, message)
        else:
            self.diffConfig.emit(NetCmd.MDF_CHN_VK, message)

        return True

    def mdf_delvk(self, name):
        """Удалить указанный ВК из конфигурации"""
        logging.debug("Delete VK: %s" % name)
        # MDF_DEL_VK

        message = '{"vkname":"' + name + '"}'

        if self._connected:
            self._client.net_request(NetCmd.MDF_DEL_VK, message)
        else:
            self.diffConfig.emit(NetCmd.MDF_DEL_VK, message)

        return True

    def mdf_addtest(self, hostname, testname, dump=False):
        """Добавить новый тест в текущую конфигурацию"""
        logging.debug("Add test: %s in %s" % (testname, hostname))
        # MDF_ADD_TEST

        t = None
        if dump:
            t = test(json.loads(dump))
            t.testname = str(testname)
        else:
            t = test()
            t.testname = str(testname)

        message = '{"vkname":"' + hostname + '",' + \
                  '"testname":"' + testname + '",' \
                                              '"test":[' + t.dump(compact=True) + ']}'

        if self._connected:
            self._client.net_request(NetCmd.MDF_ADD_TEST, message)
        else:
            self.diffConfig.emit(NetCmd.MDF_ADD_TEST, message)

        return True

    def mdf_chntest(self, hostname, testname, dump):
        """Изменить параметры указанного теста"""
        logging.debug("Change test: %s in %s" % (testname, hostname))
        # MDF_CHN_TEST

        t = test(json.loads(dump))

        message = '{"vkname":"' + hostname + '",' + \
                  '"testname":"' + testname + '",' \
                                              '"test":[' + t.dump(compact=True) + ']}'

        if self._connected:
            self._client.net_request(NetCmd.MDF_CHN_TEST, message)
        else:
            self.diffConfig.emit(NetCmd.MDF_CHN_TEST, message)

        return True

    def mdf_deltest(self, hostname, testname):
        """Удалить указанный тест из конфигурации"""
        logging.debug("Delete test: %s in %s" % (testname, hostname))
        # MDF_DEL_TEST

        message = '{"vkname":"' + hostname + '",' + \
                  '"testname":"' + testname + '"' + \
                  '}'

        if self._connected:
            self._client.net_request(NetCmd.MDF_DEL_TEST, message)
        else:
            self.diffConfig.emit(NetCmd.MDF_DEL_TEST, message)

        return True

    def mdf_chnpref(self, dump):
        """Изменить параметры глобальных параметров"""
        logging.debug("Change global preferences")
        # MDF_CHN_PREF

        message = '{"options":' + dump + '}'

        if self._connected:
            self._client.net_request(NetCmd.MDF_CHN_PREF, message)
        else:
            self.diffConfig.emit(NetCmd.MDF_CHN_PREF, message)

        return True

    def mdf_vkon(self, on=True, vk=None):
        """Включить/Выключить для тестирования указанный/все ВК
        :param on:
        :param vk:
        """
        # MDF_VK_ON     - on = True,  vk = {VK}
        # MDF_VK_OFF    - on = False, vk = {VK}
        # MDF_VKALL_ON  - on = True,  vk = None
        # MDF_VKALL_OFF - on = False, vk = None
        # TODO: MDF_VKALL_ON, MDF_VKALL_OFF

        message = '{"vkname":"' + vk + '"}'

        if on:
            if self._connected:
                self._client.net_request(NetCmd.MDF_VK_ON, message)
            else:
                self.diffConfig.emit(NetCmd.MDF_VK_ON, message)
        else:
            if self._connected:
                self._client.net_request(NetCmd.MDF_VK_OFF, message)
            else:
                self.diffConfig.emit(NetCmd.MDF_VK_OFF, message)

    def mdf_teston(self, on=True, test=None, vk=None):
        """Включить/Выключить для тестирования указанный/все тест(ы)
        :param on:
        :param test:
        """
        # TODO: MDF_TESTALL_ON, MDF_TESTALL_OFF
        # MDF_TEST_ON     - on = True,  test = {test},  vk = {VK}
        # MDF_TEST_OFF    - on = False, test = {test},  vk = {VK}
        # MDF_TESTALL_ON  - on = True,  test = None,    vk = None
        # MDF_TESTALL_OFF - on = False, test = None,    vk = None

        message = '{"testname":"' + test + '","vkname":"' + vk + '"}'

        if on:
            if self._connected:
                self._client.net_request(NetCmd.MDF_TEST_ON, message)
            else:
                self.diffConfig.emit(NetCmd.MDF_TEST_ON, message)
        else:
            if self._connected:
                self._client.net_request(NetCmd.MDF_TEST_OFF, message)
            else:
                self.diffConfig.emit(NetCmd.MDF_TEST_OFF, message)

# -----------------------------------------------------------------------------
# No __main__
