#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Stp Network


Usage:

self.stpnet = StpNetwork()
self.stpnet.sigIndication.connect(self.net_indication)
self.stpnet.sigError.connect(self.net_error)
self.stpnet.net_connect("localhost", "9876")
self.stpnet.net_request(<NetCmd.*>, <data>)
self.stpnet.net_disconnect()

def net_indication(self, message):
    # StpParser message
    mtype = message.get_type()
    mans = message.get_answer()
    ...

def net_error(self, errstr):
    print errstr

"""

import logging
import copy
from PyQt5 import QtNetwork
from PyQt5 import QtCore
from PyQt5.QtCore import QObject, QIODevice

# from PyQt5.QtCore import QCoreApplication

from stp_gui.Messages import StpParser, create_message

from stp_gui.Common import _

NETDEBUG = False


# -----------------------------------------------------------------------------
class NetCmd(object):
    """ Command to kernel:
    """
    SET_CFG = "SET_CFG"  # send config to kernel
    # SET_CFG         = "SET_CFG"         # send config to kernel
    SET_EXCL = "SET_EXCL"  # request exclusive mode
    SET_UN_EXCL = "SET_UN_EXCL"  # cancel exclusive mode
    GET_CFG = "GET_CFG"  # read config from kernel
    GET_TEST_LOG = "GET_TEST_LOG"  # request log data from kernel (about test)
    GET_INFO_ABOUT_VK = "GET_INFO_ABOUT_VK"  #
    CMD_RUN = "CMD_RUN"  # start tests
    CMD_PAUSE = "CMD_PAUSE"  # pause tests
    CMD_CONTINUE = "CMD_CONTINUE"  # resume tests
    CMD_STOP = "CMD_STOP"  # stop tests
    CMD_RESET = "CMD_RESET"  # clean received results
    MDF_ADD_VK = "MDF_ADD_VK"  # add new VK to config
    MDF_CHN_VK = "MDF_CHN_VK"  # change VK parameters
    MDF_DEL_VK = "MDF_DEL_VK"  # remove VK from config
    MDF_ADD_TEST = "MDF_ADD_TEST"  # add new Test to config
    MDF_CHN_TEST = "MDF_CHN_TEST"  # change Test parameters
    MDF_DEL_TEST = "MDF_DEL_TEST"  # remove Test from config
    MDF_CHN_PREF = "MDF_CHN_PREF"  # change global parameters
    MDF_VK_ON = "MDF_VK_ON"  # VK on
    MDF_VK_OFF = "MDF_VK_OFF"  # VK off
    MDF_VKALL_INVERT = "MDF_VKALL_INVERT"  # invert all VK
    MDF_VKALL_ON = "MDF_VKALL_ON"  # all VK on
    MDF_VKALL_OFF = "MDF_VKALL_OFF"  # all VK off
    MDF_TEST_ON = "MDF_TEST_ON"  # Test on
    MDF_TEST_OFF = "MDF_TEST_OFF"  # Test off
    MDF_TESTALL_INVERT = "MDF_TESTALL_INVERT"  # invert all Tests
    MDF_TESTALL_ON = "MDF_TESTALL_ON"  # all Tests on
    MDF_TESTALL_OFF = "MDF_TESTALL_OFF"  # all Tests off
    WRK_TERMINATE = "WRK_TERMINATE"  # terminate kernel
    WRK_DISCONNECT = "WRK_DISCONNECT"  # disconnect from all monitor
    WRK_TERMALL = "WRK_TERMALL"  # disconnect from monitors and kernel
    WRK_RECONNECT = "WRK_RECONNECT"  # reconnect to all monitor
    # MES_STATUS      = "MES_STATUS"      # current status (UNLY FROM KERNEL TO UI)
    # MES_SERVER_DONE = "MES_SERVER_DONE" # kernel exit (UNLY FROM KERNEL TO UI)
    MES_MESSAGE = "MES_MESSAGE"  # Message from kernel
    MES_ACKNOWLEDGE = "MES_ACKNOWLEDGE"  # connect to kernel
    MES_UI_DONE = "MES_UI_DONE"  # UI terminated message

    @staticmethod
    def commands():
        commands = [attr for attr in dir(NetCmd) if not callable(attr) and
                    (attr.startswith("MDF") or
                     attr.startswith("SET_CFG") or
                     attr.startswith("GET") or
                     attr.startswith("CMD"))]
        return commands


class NetAns(object):
    """ Status from kernel:
    """
    RPL_OK = "RPL_OK"  # success
    RPL_BUSY = "RPL_BUSY"  # kernel busy, try again
    RPL_NONE = "RPL_NONE"  # no reply data
    RPL_ERROR = "RPL_ERROR"  # error or wrong request
    RPL_NO_PERM = "RPL_NO_PERM"  # permission denied
    RPL_CRC_DIFF = "RPL_CRC_DIFF"  #


# -----------------------------------------------------------------------------
class StpNetwork(QtNetwork.QTcpSocket):
    """
    Network connetction to kernel
    Emit signals:
        sigIndication(StpParser) - received message from kernel
        sigError(str) - network error
    """

    # signals:
    sigIndication = QtCore.pyqtSignal(StpParser)
    sigError = QtCore.pyqtSignal(str)

    # singleton
    def __new__(cls, *args, **kw):
        if not hasattr(cls, '_instance'):
            cls._instance = super(StpNetwork, cls).__new__(cls, *args, **kw)
        return cls._instance

    def __init__(self, parent):
        super().__init__()

        # Shows if there are non acknowledged config modification queries
        self._no_ack = False
        self._timeout = 5 * 1000
        self._hostname = None
        self._port = 0
        self._curcmd = None
        self._disconnecting = False
        self._stpparser = StpParser()
        self.readyRead.connect(self.SlotReadData)
        self.error.connect(self.slotSocketError)
        self.disconnected.connect(self.slotSocketDisconnect)
        # QObject.connect(self, QtCore.pyqtSignal("readyRead()"), self, QtCore.PYQT_SLOT("slotReadData()"))
        # QObject.connect(self, SIGNAL("error()"), self, SLOT("slotSocketError()"))
        self.sync_list = NetCmd.commands()
    @QtCore.pyqtSlot()
    def SlotReadData(self):
        while self.bytesAvailable() > 40:
            # QObject.disconnect(self, SIGNAL("readyRead()"), self, SLOT("slotReadData()"))
            header = self.peek(100)
            headerStr = str(header, 'utf-8')
            header_list = headerStr.split("\n")
            while not header_list[0].startswith('STP'):
                # TODO: Maybe we should clean up in this case or panic
                self.read(1)
                header = self.peek(100)
                headerStr = str(header, 'utf-8')
                header_list = headerStr.split("\n")
                # return
            size_field = header_list[1].split(' ')
            if size_field[0] != 'Size:':
                size_field = header_list[2].split(' ')
            if size_field[0] == 'Size:':
                sz = int(size_field[1])
                if self.bytesAvailable() < sz:
                    return
                h = headerStr.split('\r\n\r\n')[0]
                mes = self.peek(len(h) + sz + 4)
                m1 = self.read(len(h) + sz + 4)
                m = str(m1, 'utf-8')
                if m:
                    m = m.split('\x00')
                    #print("slot read data 1" + m[0])
                    self._process_message(m[0])
            else:
                # If we are here, than there is a CMD_* or SET_* message
                m = headerStr.split('\r\n\r\n')[0]
                mes1 = self.read(len(m) + 4)
                mes = str(mes1, 'utf-8')
                if mes:
                   # print("slot read data 2" + mes)
                   # mes = mes.split('\x00')
                    self._process_message(mes)

    @QtCore.pyqtSlot()
    def slotSocketError(self, sockerror=None):
        self._process_error(sockerror)
        # print 'Net Error:', self.error(), self._curcmd

    @QtCore.pyqtSlot()
    def slotSocketDisconnect(self):
        if not self._disconnecting:
            mes = _("Kernel closed connection")
            self.sigError.emit(mes)

    # ------------------------------------
    #  socket's callbak:

    def _process_message(self, message):
        """ parse received message
        """
        # TODO: try exception from Messages
        if self._stpparser.add_message(message):
            mes = copy.deepcopy(self._stpparser)
            # print 'received ', mes.get_type(), mes.get_answer()
            # print 'CALL: _process_message'
            # print 'MES: ', mes.get_type(), mes.get_multi_number(), '/', mes.get_multi(), mes.get_answer()
            if mes.get_type() in self.sync_list:
                self._no_ack = False
            # received reply for current request
            if mes.get_type() == self._curcmd:
                logging.debug("NET: reply(%s) - %s" % (self._curcmd, mes.get_answer()))
                self._curcmd = None
                self._no_ack = False
            # else:
            # logging.error("Expected {}, but received {}".format(self._curcmd, mes.get_type()))
            if mes.get_multi_number() == mes.get_multi():
                self.sigIndication.emit(mes)
        else:
            pass
        if NETDEBUG:
            self._stpparser.print_message()
        # print 'EXIT: _process_message'

    def _process_error(self, socketError, errstr="Unknown"):
        """ parse socket error
        """
        message = None
        if socketError == QtNetwork.QAbstractSocket.HostNotFoundError:
            message = _("Could not resolve hostname/port: Name or service not known.")
        elif socketError == QtNetwork.QAbstractSocket.ConnectionRefusedError:
            message = _("Unable to connect to remote host: Connection refused.")
        elif socketError == None:
            # We assume that this error doesn't need to be logged.
            # It's needed to surpress following error:
            # TypeError: slotSocketError() takes exactly 2 arguments (1 given)
            return
        else:
            message = "The following error occurred: %s." % errstr
        self.sigError.emit(message)

    # ------------------------------------
    # properties:

    def _get_timeout(self):
        return self._timeout

    def _set_timeout(self, timeout):
        self._timeout = timeout

    def _get_hostname(self):
        return self._hostname

    def _set_hostname(self, hostname):
        pass

    def _get_port(self):
        return self._port

    def _set_port(self, port):
        pass

    timeout = property(_get_timeout, _set_timeout, doc='Get/set the socket timeout')
    hostname = property(_get_hostname, _set_hostname, doc='read only property socket address')
    port = property(_get_port, _set_port, doc='read only property socket port')

    # ------------------------------------
    # public:

    def net_connect(self, hostname, port):
        logging.info(_("Connect attempt, address [{0}] port [{1}]").format(hostname, port))
       # self.connectToHost(hostname, port)
        self.connectToHost(hostname, port, QIODevice.ReadWrite)
        self.waitForConnected(self._timeout)
            #self._process_error(self.error(), self.errorString())
            #return False
        self._hostname = hostname
        self._port = port
        logging.info(_("Connected to {0}:{1}").format(self._hostname, str(self._port)))
        return True

    def net_disconnect(self):
        self._disconnecting = True
        self.disconnectFromHost()
        logging.info(_("Disconnected from {0}:{1}").format(self._hostname, str(self._port)))
        self._curcmd = None
        self._disconnecting = False
        self._no_ack = False

    def net_control(self, netcmd, data=None):
        """
        request and no acknowledge
        :param netcmd: NetCmd.*
        :param data: str json or None
        """
        logging.debug("NET: request(%s)N" % (netcmd))
        self._net_sendmsg(create_message(netcmd, data))
        return True

    def net_request(self, netcmd, data=None):
        """
        request and save code for check acknowledge
        :param netcmd: NetCmd.*
        :param data: str json or None
        """
        if netcmd in self.sync_list:
            # while self._no_ack:
            # print 'a2'
            # self.waitForReadyRead(1)
            # QCoreApplication.processEvents()
            self._no_ack = True
        if self._curcmd:
            # logging.error("NET: request(%s) processed" % (netcmd))
            # return False
            pass
        logging.debug("NET: request(%s)A" % (netcmd))
        self._curcmd = netcmd
        self._net_sendmsg(create_message(netcmd, data))

        # while self._no_ack:
        # self.waitForReadyRead(1)
        # QCoreApplication.processEvents()

        return True

    # ------------------------------------

    def _net_sendmsg(self, message):
        mes = ''
        for i in message:
            mes = mes + i
        if NETDEBUG:
            print(">>>>>>>>>> REQUEST >>>>>>>>>>")
            print(str(mes))
            print("<<<<<<<<<< REQUEST <<<<<<<<<<")

        BytearrayMes = bytearray(mes, 'utf-8')
        self.writeData(BytearrayMes)
        self.flush()
